"""
Management of slicing tuples and indexing coordinates for tensors.
"""

import torch


def coordinates_to_slicing(
    minc: torch.Tensor,
    maxc: torch.Tensor,
    stride: torch.Tensor = None) -> tuple:
    """
    Converts min, max, and stride coordinates to a slicing tuple.

    Args:
        minc (Tensor): The minimum integer coordinate.
        maxc (Tensor): The maximum integer coordinate (inclusive).
        stride (Tensor, optional): The sampling strides (or steps).

    Returns:
        tuple: Slicing tuple.
    """
    if stride is None:
        stride = [None for _ in minc]
    make = lambda a, b, c: slice(int(a), int(b + 1), c if c is None else int(c))
    return tuple([make(*params) for params in zip(minc, maxc, stride)])


def expand_slicing(slicing: tuple, length: int) -> tuple:
    """
    Expands a slicing tuple to specified length, accounting for
    missing items and ellipses.

    Args:
        slicing (tuple): The original slicing tuple.
        length (int): The target slicing length.

    Returns:
        tuple: The expanded slicing.
    """
    if Ellipsis in slicing:
        # if an ellipsis is in the slicing, we need to be smart
        # about how this expands given the target length
        expanded = [slice(None) for _ in range(length)]
        for i, s in enumerate(slicing):
            if s == Ellipsis:
                if i < len(slicing) - 1:
                    remaining = slicing[i+1:]
                    expanded[-len(remaining):] = remaining
                break
            else:
                expanded[i] = s
        expanded = tuple(expanded)

        # its impossible to evaluate the slicing when multiple ellipsis are used
        if Ellipsis in expanded:
            raise ValueError('cannot use more than one ellipsis when indexing')

    else:
        # fill in any remaining elements with open slices
        missing = length - len(slicing)
        if missing > 0:
            remaining = [slice(None) for _ in range(missing)]
            expanded = (*slicing, *remaining)
        elif missing < 0:
            raise ValueError(f'slicing dimensions {len(slicing)} exceed '
                             f'limit of {length}')
        else:
            expanded = slicing

    return expanded


def slicing_to_coordinates(slicing: tuple, shape: tuple) -> tuple:
    """
    Converts a slicing tuple to min, max (inclusive), and stride coordinates.

    Args:
        slicing (tuple): The slicing tuple.
        shape (tuple of int): The shape of the tensor.

    Returns:
        tuple of Tensor: Minimum coordinates, maximum coordinates, and strides.
    """
    shape = torch.tensor(shape, dtype=torch.int)

    # make sure slicing is conformed to the right size
    slicing = expand_slicing(slicing, len(shape))

    # extract the minimum coordinate based on the slice start index
    minc = torch.zeros(len(shape), dtype=torch.int)
    for i, s in enumerate(slicing):
        if isinstance(s, slice) and s.start is not None:
            minc[i] = s.start
        elif isinstance(s, (int, torch.Tensor)):
            minc[i] = s

    # wrap any negative coordinates
    negative = minc < 0
    minc[negative] = shape[negative] + minc[negative]

    # extract the minimum coordinate based on the slice
    # stop index (subtracted by 1)
    maxc = shape - 1
    for i, s in enumerate(slicing):
        if isinstance(s, slice) and s.stop is not None:
            maxc[i] = s.stop - 1
        elif isinstance(s, (int, torch.Tensor)):
            maxc[i] = s

    # again, wrap any negative coordinates
    negative = maxc < 0
    maxc[negative] = shape[negative] + maxc[negative]

    # extract stride (if exists) from the slice step
    stride = [s.step if isinstance(s, slice) else None for s in slicing]
    stride = torch.tensor([x or 1 for x in stride]) if any(stride) else None

    return (minc, maxc, stride)


def conform_coordinates(coords: torch.Tensor, num: int = None) -> torch.Tensor:
    """
    Conform coordinates to certain shape and dimensionality.

    Args:
        coords (Tensor): Coordinate tensor of size $(1,)$ or $(3,)$ or $(3, N)$.
        num (int, optional): Number of coordinates $N$ represented in the second dimension
            of the output tensor. If None, the output tensor will have shape $(3,)$.
    
    Returns:
        Tensor: Coordinates of shape $(3, N)$ or $(3,)$.
    """
    coords = torch.as_tensor(coords)
    if coords.ndim == 0:
        coords = coords.repeat((3, num)) if num is not None else coords.repeat(3)
    elif coords.shape == (3,) and num is not None:
        coords = coords.unsqueeze(1).repeat(1, 2)
    if num is not None and coords.shape != (3, num):
        raise ValueError(f'tensor must be of size (3, {num}) or (3,) or (1,), got {coords.shape}')
    if num is None and coords.shape != (3,):
        raise ValueError(f'tensor must be of size (3,) or (1,), got {coords.shape}')
    return coords



"""
Triangular mesh representation and manipulation.
"""

from __future__ import annotations

import os
import torch
import numpy as np
import voxel as vx


class Mesh:

    def __init__(self, vertices: torch.Tensor, faces: torch.Tensor) -> None:
        """
        Triangular mesh topology in 3D world space.

        Args:
            vertices (Tensor): Vertex coordinates of shape (V, 3).
            faces (Tensor): Triangular face integer indices of shape (F, 3).
        """
        vx.caching.init_property_cache(self)
        self.vertices = vertices
        self.faces = faces

    @property
    def vertices(self) -> torch.Tensor:
        """
        Point positions represented by a (V, 3) tensor.
        """
        return self._vertices

    @vertices.setter
    def vertices(self, vertices: torch.Tensor):
        if vertices.ndim != 2 or vertices.shape[-1] != 3:
            raise ValueError('expected shape (V, 3) for vertex tensor, '
                             f'but got {vertices.shape}')
        self._vertices = vertices

    @property
    def num_vertices(self) -> int:
        """
        Total number of vertices in the mesh.
        """
        return len(self.vertices)

    @property
    def faces(self) -> torch.Tensor:
        """
        Triangle faces represented by a (F, 3) tensor.
        """
        return self._faces

    @faces.setter
    def faces(self, faces: torch.Tensor):
        if faces.ndim != 2 or faces.shape[-1] != 3:
            raise ValueError(f'expected shape (F, 3) for faces array, but got {faces.shape}')
        self._faces = faces.int()

    @property
    def num_faces(self) -> int:
        """
        Total number of faces in the mesh.
        """
        return len(self.faces)

    def new(self, vertices: torch.Tensor) -> Mesh:
        """
        Construct a new mesh instance with the provided vertices, while
        preserving any unchanged properties of the original.

        Args:
            vertices (Tensor): The new vertices tensor replacement.
        """
        mesh = self.__class__(vertices, self.faces)
        vx.caching.transfer_property_cache(self, mesh)
        return mesh

    @property
    def device(self) -> torch.Device:
        """
        The device of the mesh vertex and face tensors.
        """
        return self.vertices.device

    def to(self, device: torch.Device) -> Mesh:
        """
        Move the mesh (vertex and faces tensors) to a device.

        Args:
            device: A torch device.

        Returns:
            Mesh: A new mesh instance.
        """
        if device is None:
            return self
        return Mesh(self.vertices.to(device), self.faces.to(device))

    def cpu(self) -> Mesh:
        """
        Move the mesh vertex and face tensors to the CPU.

        Returns:
            Mesh: A new mesh instance with with the data on the CPU.
        """
        return Mesh(self.vertices.cpu(), self.faces.cpu())

    def cuda(self) -> Mesh:
        """
        Move the mesh vertex and face tensors to the GPU.

        Returns:
            Mesh: A new mesh instance with with the data on the GPU.
        """
        return Mesh(self.vertices.cuda(), self.faces.cuda())

    def type(self, dtype: torch.dtype) -> Mesh:
        """
        Cast the mesh vertices to a new data type.

        Args:
            dtype (torch.dtype): The target vertex data type.

        Returns:
            Mesh: A new mesh instance with the casted vertices.
        """
        return self.new(self.vertices.type(dtype))

    def save(self, filename: os.PathLike, fmt: str = None, **kwargs) -> None:
        """
        Save the mesh to a file.

        Args:
            filename (PathLike): The path to the file to save.
            fmt (str, optional): The format of the file. If None, the format is
                determined by the file extension.
            kwargs: Additional arguments passed to the file writing method.
        """
        vx.save_mesh(self, filename, fmt=fmt, **kwargs)

    @vx.caching.cached
    def triangles(self) -> torch.Tensor:
        """
        Triangle coordinate arrary with shape (F, 3, 3).
        """
        return self.vertices[self.faces]

    @vx.caching.cached
    def triangles_cross(self) -> torch.Tensor:
        """
        Vertex cross-product with shape (F, 3).
        """
        vecs = torch.diff(self.triangles, dim=1)
        return torch.cross(vecs[:, 0], vecs[:, 1], dim=1)

    @vx.caching.cached_transferable
    def edges(self) -> torch.Tensor:
        """
        All directional edges in the mesh, with shape (E, 2). Note these are not unique.
        """
        return self.faces.repeat_interleave(2, dim=-1).roll(-1, dims=-1).view((-1, 2))

    @vx.caching.cached_transferable
    def uniform_laplacian(self) -> torch.Tensor:
        """
        The sparse uniform Laplacian matrix for the mesh connectivity.
        Note that it is generally faster to use the `gather()` or `smooth_features()`
        methods for diffusing features on the mesh graph.
        """
        V = self.num_vertices

        # compute the uniform weight matrix
        dest = self.edges[:, 0]
        count = torch.bincount(dest, minlength=V)
        weight = count[dest]
        val = torch.where(weight > 0.0, 1.0 / weight, weight)
        L = torch.sparse_coo_tensor(self.edges.T.type(torch.int64), val, (V, V))

        # subtract identity
        idx = torch.arange(V, device=self.device)
        idx = torch.stack([idx, idx], dim=0)
        # NOTE: once sparse torch.eye is implemented globally, use that instead
        ones = torch.ones(V, dtype=torch.float32, device=self.device)
        L -= torch.sparse_coo_tensor(idx, ones, (V, V))
        return L

    @vx.caching.cached_transferable
    def edge_face(self) -> torch.Tensor:
        """
        Face indices corresponding to each directional edge in the
        mesh, with shape (E,).
        """
        arange = torch.arange(self.num_faces, device=self.faces.device)
        return arange.repeat_interleave(3).reshape(-1)

    @vx.caching.cached_transferable
    def unique_edge_indices(self) -> tuple:
        """
        Indices that extract all unique edges from the directional edge list.
        """
        aligned = self.edges.sort(dim=1)[0]

        # this is a way to do lexsort in pytorch, which is faster
        # than using torch.unique
        idx = aligned[:, 1].argsort(dim=-1, stable=True)
        order = idx.gather(-1, aligned[:, 0].gather(-1, idx).argsort(dim=-1, stable=True))

        pef = aligned[order]
        device = self.faces.device
        shift = torch.cat([torch.tensor([True], device=device),
                           torch.any(pef[1:] != pef[:-1], dim=-1),
                           torch.tensor([True], device=device)])
        matched = shift.nonzero(as_tuple=False).squeeze(-1)
        repeated = torch.repeat_interleave(torch.arange(len(matched) - 1, device=device),
                                           (matched[1:] - matched[:-1]))
        reverse = repeated[order.argsort()]
        indices = order[matched[:-1]]

        repeats = torch.all(pef[:-1] == pef[1:], dim=-1)
        matched = repeats.nonzero(as_tuple=False).squeeze(-1)
        bidir_indices = order[torch.stack([matched, matched + 1], dim=1)]

        return indices, reverse, bidir_indices

    @vx.caching.cached_transferable
    def unique_edges(self):
        """
        Unique bi-directional edges in the mesh, with shape (U, 2).
        """
        return self.edges[self.unique_edge_indices[0]]

    @vx.caching.cached_transferable
    def adjacent_faces(self):
        """
        Adjacent faces indices corresponding to each edge in `unique_edges`.
        """
        return self.edge_face[self.unique_edge_indices[2]]

    @vx.caching.cached
    def face_normals(self) -> torch.Tensor:
        """
        Face (unit) normals with shape (F, 3).
        """
        return torch.nn.functional.normalize(self.triangles_cross)

    @vx.caching.cached
    def face_areas(self) -> torch.Tensor:
        """
        Face areas with shape (F,)
        """
        return (self.triangles_cross ** 2).sum(-1).sqrt() / 2

    @vx.caching.cached
    def face_angles(self) -> torch.Tensor:
        """
        Face angles (in radians) with shape (F, 3).
        """
        triangles = self.triangles
        u = torch.nn.functional.normalize(triangles[:, 1] - triangles[:, 0])
        v = torch.nn.functional.normalize(triangles[:, 2] - triangles[:, 0])
        w = torch.nn.functional.normalize(triangles[:, 2] - triangles[:, 1])
        angles = torch.zeros((self.num_faces, 3), device=u.device)
        angles[:, 0] = ( u * v).sum(-1).clamp(-1, 1).arccos()
        angles[:, 1] = (-u * w).sum(-1).clamp(-1, 1).arccos()
        angles[:, 2] = torch.pi - angles[:, 0] - angles[:, 1]
        return angles

    @vx.caching.cached
    def vertex_normals(self) -> torch.Tensor:
        """
        Vertex (unit) normals, computed from face normals weighted by their angle.
        """
        scalars = self.face_angles.view(-1, 1) * self.face_normals.repeat_interleave(3, dim=0)
        indices = self.faces.type(torch.int64).view(-1, 1).expand(-1, 3)
        normals = torch.zeros_like(self.vertices).scatter_add(-2, indices, scalars)
        return torch.nn.functional.normalize(normals)

    @vx.caching.cached
    def vertex_areas(self) -> torch.Tensor:
        """
        The total face surface area contributed to each vertex.
        """
        area_contributions = self.face_areas.unsqueeze(-1) * self.face_angles / torch.pi
        source = area_contributions.view(-1)
        indices = self.faces.view(-1).long()
        zeros = torch.zeros(self.num_vertices, dtype=self.vertices.dtype, device=self.device)
        return zeros.scatter_reduce(0, indices, source, reduce='sum', include_self=False)

    def flip_faces(self) -> vx.Mesh:
        """
        Flip triangular face directions.
        """
        return vx.Mesh(self.vertices, self.faces.flip(-1))

    def gather(self, features: torch.Tensor, reduce: str = 'mean') -> torch.Tensor:
        """
        """
        edges = self.edges
        source = features[edges[:, 0]]
        indices = edges[:, 1].type(torch.int64).view(-1, 1).expand(-1, source.shape[-1])
        reduced = torch.zeros_like(features).scatter_reduce(-2, indices, source,
                                                            reduce=reduce, include_self=False)
        return reduced

    def smooth_mesh(self, alpha: float = 0.5, iterations: int = 1) -> Mesh:
        """
        Smooth the mesh vertex positions with the uniform Laplacian operator.

        Args:
            alpha (float, optional): Smoothing factor between 0 and 1. Defaults to 0.5.
            iterations (int, optional): Number of smoothing iterations. Defaults to 1.

        Returns:
            Mesh: Smoothed mesh.
        """
        return self.new(self.smooth_features(self.vertices, alpha=alpha, iterations=iterations))

    def smooth_features(self, features: torch.Tensor, alpha: float = 0.5, iterations: int = 1) -> torch.Tensor:
        """
        Smooth vertex features with the uniform Laplacian operator. Note that this does not actually use
        the sparse Laplacian matrix, but rather the vertex gather method, which is roughly 2x faster.

        Args:
            features (Tensor): Input features to smooth of the shape $(V,)$ or $(V, C)$.
            alpha (float, optional): Smoothing factor between 0 and 1. Defaults to 0.5.
            iterations (int, optional): Number of smoothing iterations. Defaults to 1.

        Returns:
            Tensor: Smoothed features matching the input shape.
        """
        is1d = features.ndim == 1
        if is1d:
            features = features.view(-1, 1)
        for _ in range(iterations):
            features = (1 - alpha) * features + alpha * self.gather(features)
        if is1d:
            features = features.squeeze(-1)
        return features

    def transform(self, transform: vx.AffineMatrix) -> Mesh:
        """
        Transform mesh vertex coordinates with an affine matrix.

        Args:
            transform (AffineMatrix): Affine transformation.

        Returns:
            Mesh: Transformed mesh.
        """
        return Mesh(transform.transform(self.vertices), self.faces)

    def bounds(self, margin: float | torch.Tensor = None) -> Mesh:
        """
        Compute a box mesh enclosing the vertex bounds.

        Args:
            margin (float or Tensor, optional): Margin (in vertex units) to expand
                the cropping boundary. Can be a positive or negative delta.

        Returns:
            Mesh: Bounding box mesh.
        """
        min_point = self.vertices.amin(dim=0).float()
        max_point = self.vertices.amax(dim=0).float()

        # expand (or shrink) margin around border
        if margin is not None:
            margin = vx.slicing.conform_coordinates(margin, 2)
            min_point -= margin[:, 0]
            max_point += margin[:, 1]

        return construct_box_mesh(min_point, max_point)

    def extract_submesh(self, vertex_mask: torch.Tensor) -> Mesh:
        """
        Extract a submesh containing only the vertices included in an input mask.

        Args:
            vertex_mask (Tensor): A boolean vertex mask indicating which vertices
                to include in the submesh.

        Returns:
            Mesh: Extracted mesh with a subset of vertices and remapped face indices.
        """
        vertex_mask = vertex_mask.bool()
        new_num_vertices = vertex_mask.count_nonzero()

        mapping = torch.full((self.num_vertices,), -1, dtype=torch.int32, device=vertex_mask.device)
        mapping[vertex_mask] = torch.arange(new_num_vertices, dtype=torch.int32, device=vertex_mask.device)

        faces = mapping[self.faces]
        faces = faces[(faces >= 0).all(dim=1)]

        return Mesh(self.vertices[vertex_mask], faces)

    def largest_connected_components(self, k: int = 1) -> torch.Tensor:
        """
        Compute a mask indicating the top k largest connected components in the mesh graph, where
        'largest' is defined by number of vertices. This method uses scipy under the hood and
        therefore runs on the CPU.

        Args:
            k (int, optional): The top k largest components to include. Defaults to 1.

        Returns:
            Tensor: A boolean vertex mask representing vertices in the top k components.
        """
        import scipy.sparse as sp

        # convert to a sparse scipy matrix and compute the components
        N = self.num_vertices
        row, col = self.edges.detach().cpu().T.numpy()
        adj = sp.coo_matrix((np.ones(len(row)), (row, col)), (N, N))
        num_found, components = sp.csgraph.connected_components(adj)

        if num_found <= k:
            return torch.ones(self.num_vertices, device=self.vertices.device, dtype=torch.bool)

        # compile the top largest into a single vertex mask
        _, count = np.unique(components, return_counts=True)
        subset_np = np.in1d(components, count.argsort()[-k:])
        vertex_mask = torch.from_numpy(subset_np).to(device=self.vertices.device, dtype=torch.bool)
        return vertex_mask


def construct_box_mesh(min_point: torch.Tensor, max_point: torch.Tensor) -> Mesh:
    """
    Construct a rectangular box mesh from a lower and upper corner coordinate.

    Args:
        min_point (Tensor): Coordinate at the lower corner bound.
        max_point (Tensor): Coordinate at the upper corner bound.

    Returns:
        Mesh: Rectangular mesh.
    """
    mask = torch.tensor([
        [0, 0, 0],
        [1, 0, 0],
        [1, 1, 0],
        [0, 1, 0],
        [0, 0, 1],
        [1, 0, 1],
        [1, 1, 1],
        [0, 1, 1]], device=min_point.device)

    faces = torch.tensor([
        [0, 1, 2],
        [0, 2, 3],
        [4, 5, 6],
        [4, 6, 7],
        [0, 1, 5],
        [0, 5, 4],
        [2, 3, 7],
        [2, 7, 6],
        [0, 3, 7],
        [0, 7, 4],
        [1, 2, 6],
        [1, 6, 5]], device=min_point.device)

    points = (mask == 0) * (min_point - 0.5) + mask * (max_point + 0.5)
    return Mesh(points, faces)



"""
Bounding box utilities.
"""

from __future__ import annotations

import pathlib
import torch
import voxel as vx


class BoundingBox:

    def __init__(self,
        center: torch.Tensor = None,
        rotation: torch.Tensor = None,
        extent: torch.Tensor = None):
        """
        Args:
            center (torch.Tensor): Center of the bounding box.
            rotation (torch.Tensor): Rotation of the bounding box.
            extent (torch.Tensor): Extents of the bounding box.
        """
        self.center = torch.zeros(3) if center is None else center
        self.rotation = torch.eye(3) if rotation is None else rotation
        self.extent = torch.ones(3) if extent is None else extent

    def to(self, device: torch.Device) -> BoundingBox:
        """
        Move all bounding box parameters to a device.

        Args:
            device: A torch device.

        Returns:
            BoundingBox: A new bounding box instance.
        """
        return BoundingBox(self.center.to(device), self.rotation.to(device), self.extent.to(device))

    def save(self, filename: pathlib.Path) -> None:
        """
        Save the bounding box to file.

        Args:
            filename (Path): File destination.
        """
        params = dict(center=self.center, rotation=self.rotation, extent=self.extent)
        torch.save(params, filename)

    def corner_points(self) -> torch.Tensor:
        """
        Compute the corner points of the bounding box.

        Returns:
            Tensor: Corner point tensor of shape (8, 3).
        """
        signs = torch.tensor([
            [0, 0, 0],
            [1, 0, 0],
            [1, 1, 0],
            [0, 1, 0],
            [0, 0, 1],
            [1, 0, 1],
            [1, 1, 1],
            [0, 1, 1]],
            dtype=torch.float32,
            device=self.center.device) * 2 - 1

        T = torch.eye(4, device=self.center.device, dtype=self.center.dtype)
        T[:3, :3] = self.rotation @ torch.diag(self.extent)
        T[:3, 3] = self.center
        affine = vx.AffineMatrix(T)

        return affine.transform(signs)

    def min_max_coords(self) -> tuple[torch.Tensor, torch.Tensor]:
        """
        Compute the minimum and maximum coordinates of the bounding box.

        Returns:
            tuple: Minimum and maximum coordinates.
        """
        points = self.corner_points()
        min_coord = points.amin(dim=0)
        max_coord = points.amax(dim=0)
        return min_coord, max_coord

    def mesh(self) -> vx.Mesh:
        """
        Construct a rectangular box mesh from the bounding box.
        
        Returns:
            Mesh: Rectangular box mesh.
        """
        vertices = self.corner_points()

        # triangular faces
        faces = torch.tensor([
            [0, 1, 2],
            [0, 2, 3],
            [4, 5, 6],
            [4, 6, 7],
            [0, 4, 5],
            [0, 5, 1],
            [1, 5, 6],
            [1, 6, 2],
            [2, 6, 7],
            [2, 7, 3],
            [3, 7, 4],
            [3, 4, 0]],
            dtype=torch.int32,
            device=self.center.device)
        
        return vx.Mesh(vertices, faces)
    
    def geometry(self, spacing: torch.Tensor = None) -> vx.AcquisitionGeometry:
        """
        Construct an acquisition geometry from the bounding box.

        Args:
            spacing (torch.Tensor): Desired voxel spacing of the geometry.
        
        Returns:
            AcquisitionGeometry: Acquisition geometry.
        """
        if spacing is None:
            spacing = torch.ones(3, device=self.center.device, dtype=torch.float32)
        elif not torch.is_tensor(spacing):
            spacing = torch.tensor(spacing, device=self.center.device, dtype=torch.float32)
        if spacing.ndim == 0:
            spacing = spacing.repeat(3)
        if spacing.ndim != 1 or spacing.shape[0] != 3:
            raise ValueError(f'expected 3D spacing, got {spacing.ndim}D')

        # compute volume size
        baseshape = torch.ceil(2 * self.extent / spacing).int()

        # compute half length for re-centering
        half = spacing * (baseshape - 1) / 2

        # compute a voxel to world matrix
        T = torch.eye(4, device=self.center.device)
        T[:3, :3] = self.rotation @ torch.diag(spacing)
        T[:3, 3] = self.center - self.rotation @ half

        return vx.AcquisitionGeometry(baseshape, T)

    def expand(self, margin: float = None, factor: float = None) -> BoundingBox:
        """
        Expand the bounding box by a margin or a factor.

        Args:
            margin (float, optional): Margin to expand the bounding box.
            factor (float, optional): Factor to scale the bounding box.
        
        Returns:
            BoundingBox: Expanded bounding box.
        """
        if margin is not None:
            extent = self.extent + margin
        elif factor is not None:
            extent = self.extent * factor
        else:
            raise ValueError('either `margin` or `factor` should be provided')
        return BoundingBox(self.center, self.rotation, extent)

    def rotate(self, rotation: torch.Tensor, degrees: bool = True) -> BoundingBox:
        """
        Applies a rotation to the bounding box.

        Args:
            rotation (Tensor): Rotation angles. If `degrees` is True, the
                angles are in degrees, otherwise they are in radians.
            degrees (bool, optional): Whether the angles are defined as degrees or,
                alternatively, as radians.
        
        Returns:
            BoundingBox: Rotated bounding box.
        """
        matrix = vx.affine.angles_to_rotation_matrix(rotation, degrees)[:3, :3]
        rotated = matrix @ self.rotation
        return BoundingBox(self.center, rotated, self.extent)

    def fit_extent(self, points: torch.Tensor) -> BoundingBox:
        """
        Fit the extent of the bounding box to a set of points.

        Args:
            points (torch.Tensor): Coordinate point cloud.

        Returns:
            BoundingBox: Bounding box with refit extent.
        """
        assert points.ndim == 2 and points.shape[1] == 3, "points should be of shape (N, 3)"

        # center points and project onto eigenvectors
        centroid = points.mean(dim=0)
        centered_points = points - centroid
        projected = centered_points @ self.rotation

        # find min and max along each principal axis
        min_proj = projected.amin(dim=0)
        max_proj = projected.amax(dim=0)

        # compute OBB parameters and recompute the center in global coordinates
        center_local = (min_proj + max_proj) / 2
        extents = (max_proj - min_proj) / 2
        obb_center = centroid + self.rotation @ center_local

        return BoundingBox(obb_center, self.rotation, extents)

    def fine_tune(self, points: torch.Tensor) -> BoundingBox:
        """
        Orient the bounding box to minimize the volume of the box enclosing a point cloud.

        Args:
            points (Tensor): Coordinate point cloud of shape (N, 3).

        Returns:
            BoundingBox: Fine-tuned bounding box.
        """
        return obbox_fine_tune(points, self.rotation)


def load_bounding_box(filename: pathlib.Path) -> BoundingBox:
    """
    Load a bounding box from file.

    Args:
        filename (Path): Target file to load.
    
    returns:
        BoundingBox: Loaded bounding box.
    """
    return BoundingBox(**torch.load(filename, weights_only=False))


def obbox(points: torch.Tensor, initialize: bool = True, fine_tune: bool = True) -> BoundingBox:
    """
    Compute an oriented bounding box (OBB).
    
    Args:
        points (Tensor): Coordinate point cloud of shape (N, 3).
        initialize (bool, optional): Whether to initialize the rotation with PCA.
        fine_tune (bool, optional): Whether to fine-tune the bounds to minimize the volume.
        
    Returns:
        BoundingBox: Oriented bounding box.
    """
    assert initialize or fine_tune, 'either `initialize` or `fine-tune` should be enabled'

    bounds = obbox_pca(points) if initialize else BoundingBox()

    if fine_tune:
        rotation = bounds.rotation if initialize else None
        bounds = obbox_fine_tune(points, rotation)

    return bounds


def obbox_pca(points: torch.Tensor) -> BoundingBox:
    """
    Compute an oriented bounding box (OBB) using PCA.

    Args:
        points (Tensor): Coordinate point cloud of shape (N, 3).

    Returns:
        BoundingBox: Oriented bounding box.
    """
    assert points.ndim == 2 and points.shape[1] == 3, 'points should be of shape (N, 3)'

    # center points
    centroid = points.mean(dim=0)
    centered_points = points - centroid

    # compute covariance matrix (3x3)
    cov = centered_points.t() @ centered_points / points.shape[0]

    # eigenvalues and eigenvectors
    _, eigenvectors = torch.linalg.eigh(cov)

    # project points onto eigenvectors
    projected = centered_points @ eigenvectors

    # find min and max along each principal axis
    min_proj = projected.amin(dim=0)
    max_proj = projected.amax(dim=0)

    # compute OBB parameters
    extent = (max_proj - min_proj) / 2
    
    # compute the center in global coordinates
    center_local = (min_proj + max_proj) / 2
    obb_center = centroid + eigenvectors @ center_local

    return BoundingBox(obb_center, eigenvectors, extent)


def obbox_fine_tune(points: torch.Tensor, initial_rotation: torch.Tensor = None) -> BoundingBox:
    """
    Fine-tune an oriented bounding box (OBB) to minimize the volume of the box
    enclosing a point cloud.

    Args:
        points (Tensor): Coordinate point cloud of shape (N, 3).
        initial_rotation (Tensor, optional): Initial rotation matrix of shape (3, 3).
    
    Returns:
        BoundingBox: Oriented bounding box.
    """
    assert points.ndim == 2 and points.shape[1] == 3, 'points should be of shape (N, 3)'

    stepsize = 1e-2
    maxsteps = 200

    if initial_rotation is None:
        initial_rotation = torch.eye(3, device=points.device)

    with torch.no_grad():

        centroid = points.mean(dim=0)
        centered_points = points - centroid

        # initialize angles and center deltas
        angles = torch.zeros(3, device=points.device)
        center = torch.zeros(3, device=points.device)

        # compute initial volume for reference and as a gradient scaling factor
        projected = centered_points @ initial_rotation
        min_proj = projected.amin(dim=0)
        max_proj = projected.amax(dim=0)
        inital_volume = (max_proj - min_proj).prod()

        # optimization loop
        history = []
        for step in range(maxsteps):

            # construct rotation matrix
            zero = torch.zeros((), device=points.device)
            one = torch.ones((), device=points.device)
            cos = angles.cos()
            sin = angles.sin()
            rx = torch.stack([one, zero, zero, zero, cos[0], sin[0], zero, -sin[0], cos[0]]).view(3, 3)
            ry = torch.stack([cos[1], zero, sin[1], zero, one, zero, -sin[1], zero, cos[1]]).view(3, 3)
            rz = torch.stack([cos[2], sin[2], zero, -sin[2], cos[2], zero, zero, zero, one]).view(3, 3)
            delta_rotation = rx @ ry @ rz

            rotation = initial_rotation @ delta_rotation
            translated = centered_points + center
            projected = translated @ rotation

            min_proj, min_idx = projected.min(dim=0)
            max_proj, max_idx = projected.max(dim=0)

            diff = max_proj - min_proj
            volume = diff.prod()

            # compute first set of projected gradients
            d_max = volume / diff
            d_projected = torch.zeros_like(projected)
            d_projected[min_idx, [0, 1, 2]] = -d_max
            d_projected[max_idx, [0, 1, 2]] =  d_max

            # center gradients
            d_translated = d_projected @ rotation.T
            d_center = d_translated.sum(dim=0)
            d_rotation = translated.T @ d_projected
            d_delta_rotation = initial_rotation.T @ d_rotation

            # matrix composition gradients
            d_rx = d_delta_rotation @ rz.T @ ry.T
            d_ry = rx.T @ d_delta_rotation @ rz.T
            d_rz = (rx @ ry).T @ d_delta_rotation

            # sine gradients
            d_cos = torch.stack([d_rx[1, 1] + d_rx[2, 2], d_ry[0, 0] + d_ry[2, 2], d_rz[0, 0] + d_rz[1, 1]])
            d_sin = torch.stack([d_rx[1, 2] - d_rx[2, 1], d_ry[0, 2] - d_ry[2, 0], d_rz[0, 1] - d_rz[1, 0]])
            d_angles = d_cos * -sin + d_sin * cos
            d_angles /= inital_volume

            # update angles and center
            angles = angles - stepsize * d_angles
            center = center - stepsize * d_center

            # normalized volume relative to the initial volume
            relative_cost = volume / inital_volume

            # early stopping if flat improvement - the window was chosen somewhat arbitrarily
            window = 20
            threshold = sum(history[-window:]) / window
            if step > window and relative_cost > threshold:
                break

            history.append(relative_cost.item())

    # use the original rotation if the cost is higher    
    if relative_cost > 1:
        rotation = initial_rotation
        projected = centered_points @ rotation
        min_proj = projected.amin(dim=0)
        max_proj = projected.amax(dim=0)

    # compute OBB parameters
    center_local = (min_proj + max_proj) / 2
    obb_center = centroid + rotation @ center_local
    extent = (max_proj - min_proj) / 2

    return BoundingBox(obb_center, rotation, extent)



"""
Utilies for image grid filtering and kernel construction.
"""

from __future__ import annotations

import torch


def gaussian_kernel_1d(
    sigma: float,
    truncate: float = 2,
    device: torch.device | None = None,
    dtype: torch.dtype | None = None) -> torch.Tensor:
    """
    Generate a 1D Gaussian kernel with a specified standard deviation.

    Args:
        sigma (float): Standard deviations in element (voxel) space.
        truncate (float, optional): The number of standard deviations to extend
            the kernel before truncating.
        device (torch.device, optional): The device on which to create the kernel.
        device (torch.dtype, optional): The kernel datatype.

    Returns:
        Tensor: A kernel of shape $2 * truncate * sigma + 1$.
    """
    r = int(truncate * sigma + 0.5)
    x = torch.arange(-r, r + 1, device=device, dtype=dtype)
    sigma2 = 1 / torch.clip(torch.as_tensor(sigma), min=1e-5).pow(2)
    pdf = torch.exp(-0.5 * (x.pow(2) * sigma2))
    return pdf / pdf.sum()


def gaussian_blur(
    image: torch.Tensor,
    sigma: list,
    batched: bool = False,
    truncate: float = 2,
    stride: int | tuple[int] | None = None,
    padding: str = 'same') -> torch.Tensor:
    """
    Apply Gaussian blurring to a data grid.

    The Gaussian filter is applied using convolution. The size of the filter kernel
    is determined by the standard deviation and the truncation factor.

    Args:
        image (Tensor): An image tensor with preceding channel dimensions. A
            batch dimension can be included by setting `batched=True`.
        sigma (float): Standard deviations in element (voxel) space.
        batched (bool, optional): If True, assume image has a batch dimension.
        truncate (float, optional): The number of standard deviations to extend
            the kernel before truncating.
        stride (int | tuple[int] | None, optional): The stride of the convolution.
        padding (str, optional): Padding mode for the convolution. Default is 'same'.

    Returns:
        Tensor: The blurred tensor with the same shape as the input tensor.
    """
    ndim = image.ndim - (2 if batched else 1)

    # sanity check for common mistake
    if ndim == 4 and not batched:
        raise ValueError(f'gaussian blur input has {image.ndim} dims, '
                          'but batched option is False')

    # make sure sigmas match the ndim
    sigma = torch.as_tensor(sigma)
    if sigma.ndim == 0:
        sigma = sigma.repeat(ndim)
    if len(sigma) != ndim:
        raise ValueError(f'sigma must be {ndim}D, but got length {len(sigma)}')

    # make sure strides match the ndim
    if stride is not None:
        stride = torch.as_tensor(stride)
        if stride.ndim == 0:
            stride = stride.repeat(ndim)
        if len(stride) != ndim:
            raise ValueError(f'stride must be {ndim}D, but got length {len(stride)}')

    blurred = image.float()
    if not batched:
        blurred = blurred.unsqueeze(0)

    for dim, s in enumerate(sigma):

        # reuse previous kernel if we can
        if dim == 0 or s != sigma[dim - 1]:
            kernel = gaussian_kernel_1d(s, truncate, blurred.device, blurred.dtype)
        
        # kernels are normalized. if the length is one, there's no point in using it.
        # but we still need to apply the stride if it's greater than 1
        if len(kernel) == 1:
            if stride is not None and stride[dim] != 1:
                slicing = [slice(None)] * (ndim + 2)
                slicing[dim + 2] = slice(None, None, stride[dim])
                blurred = blurred[slicing]
            continue

        # set the stride for the current dimension
        if stride is not None:
            stride_dim = [1 if d != dim else stride[dim] for d in range(ndim)]
        else:
            stride_dim = 1

        # select the kernel for the current dimension
        slices = [None] * (ndim + 2)
        slices[dim + 2] = slice(None)
        kernel_dim = kernel[slices]

        # expand the kernel for multi-channel images
        num_channels = image.shape[0]
        if num_channels > 1:
            kernel_dim = kernel_dim.expand((num_channels, *kernel_dim.shape[1:]))

        # apply the convolution
        conv = getattr(torch.nn.functional, f'conv{ndim}d')
        blurred = conv(blurred, kernel_dim, groups=num_channels, stride=stride_dim, padding=padding)

    if not batched:
        blurred = blurred.squeeze(0)

    return blurred


def dilate(image: torch.Tensor, iterations: int = 1, batched: bool = False) -> torch.Tensor:
    """
    Apply a binary dilation operation to a data grid.

    Args:
        image (Tensor): An image tensor with preceding channel dimensions. A
            batch dimension can be included by setting `batched=True`.
        iterations (int, optional): The number of dilation iterations.
        batched (bool, optional): If True, assume image has a batch dimension.

    Returns:
        Tensor: The dilated tensor with the same shape as the input tensor.
    """
    ndim = image.ndim - (2 if batched else 1)

    # sanity check for common mistake
    if ndim == 4 and not batched:
        raise ValueError(f'dilate input has {image.ndim} dims, '
                          'but batched option is False')

    dilated = image.float()
    if not batched:
        dilated = dilated.unsqueeze(0)

    kernel = torch.zeros([3] * ndim, device=dilated.device, dtype=dilated.dtype)
    for dim in range(ndim):
        slices = [slice(None)] * ndim
        slices[dim] = 1
        kernel[tuple(slices)] = 1
    kernel = kernel.view(1, 1, *kernel.shape)

    conv = getattr(torch.nn.functional, f'conv{ndim}d')
    for _ in range(iterations):
        dilated = conv(dilated, kernel, groups=image.shape[0], padding='same')

    if not batched:
        dilated = dilated.squeeze(0)

    return (dilated > 0).to(image.dtype)



"""
Image and world coordinate space representations.
"""

from __future__ import annotations


space_lookup = {
    'voxel': 'V',
    'vox': 'V',
    'image': 'V',
    'world': 'W',
}


class Space:
    """
    An object designating either the voxel (image grid) or world coordinate space.
    """

    def __init__(self, space: Space | str) -> None:
        """
        Args:
            space (Space | str): Coordinate space. If string, can be 'voxel' or 'world'.
        """
        if isinstance(space, Space):
            self.code = space.code
        else:
            match = space_lookup.get(space)
            if match is None:
                raise ValueError(f'unknown space: {space}')
            self.code = match

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}('{self._categories[self.code]}')"

    def __eq__(self, value: object) -> bool:
        if isinstance(value, Space):
            return self.code == value.code
        elif isinstance(value, str):
            return self.code == space_lookup[value]
        else:
            raise ValueError(f'cannot cast {type(value)} to space')



__version__ = '0.1.1'

from . import caching
from . import slicing
from . import filters

from . import space
from .space import Space

from . import affine
from .affine import AffineMatrix
from .affine import AffineVolumeTransform

from . import bounds
from .bounds import BoundingBox
from .bounds import load_bounding_box

from . import acquisition
from .acquisition import AcquisitionGeometry
from .acquisition import Orientation
from .acquisition import cast_orientation
from .acquisition import cast_acquisition_geometry

from . import volume
from .volume import Volume
from .volume import volumes_equal

from . import mesh
from .mesh import Mesh

from . import io
from .io.volume import load_volume
from .io.volume import save_volume
from .io.mesh import load_mesh
from .io.mesh import save_mesh
from .io.affine import load_affine
from .io.affine import save_affine



"""
Methods related to a volumetric image grid with a world-space geometry.
"""

from __future__ import annotations

import os
import torch
import voxel as vx


class Volume:
    """
    A multi-channel volumetric (3D) image with a world-space representation.

    The volume grid has dimensions $(C, W, H, D)$ where $C$ is the number of
    feature channels and $W, H, D$ are the spatial width, height, and depth
    of the image (called the **baseshape**).
    """

    def __init__(self,
        tensor: torch.Tensor,
        geometry: vx.AcquisitionGeometry | vx.AffineMatrix = None) -> None:
        """
        Args:
            tensor (Tensor): Image data tensor of shape $(C, W, H, D)$ or $(W, H, D)$. 
            geometry (AcquisitionGeometry or AffineMatrix, optional): Affine geometry
                or matrix representing the voxel-to-world coordinate transform. If
                None, it defaults to a shifted identity in which the image volume
                is centered at the world origin.
        """
        if tensor.ndim == 3:
            tensor = tensor.unsqueeze(0)
        elif tensor.ndim != 4:
            raise ValueError(f'expected 3D or 4D features, got a {tensor.ndim}D input')
        self._tensor = tensor
        self.geometry = geometry

    # -------------------------------------------------------------------------
    # property getters and setters and core methods
    # -------------------------------------------------------------------------

    @property
    def tensor(self) -> torch.Tensor:
        """
        The volume feature tensor, always of shape $(C, W, H, D)$.
        """
        return self._tensor

    @property
    def geometry(self) -> vx.AcquisitionGeometry:
        """
        The acquisition geometry representing the transformation from
        voxel-center coordinates to world-space (or scanner) coordinates.
        """
        return self._geometry

    @geometry.setter
    def geometry(self, geometry: vx.AcquisitionGeometry):
        if not isinstance(geometry, vx.AcquisitionGeometry):
            geometry = vx.AcquisitionGeometry(self.baseshape, matrix=geometry, device=self.device)
        elif geometry.baseshape != self.baseshape:
            raise ValueError(f'acquisition geometry shape {tuple(geometry.baseshape)} must '
                             f'match the image base shape {tuple(self.baseshape)}')
        self._geometry = geometry

    @property
    def shape(self) -> torch.Size:
        """
        The 4D $(C, W, H, D)$ shape of the volume, including channel dimension.
        """
        return self.tensor.shape

    @property
    def baseshape(self) -> torch.Size:
        """
        The spatial 3D $(W, H, D)$ shape of the volume, excluding channel dimension.
        """
        return self.tensor.shape[1:]

    @property
    def num_channels(self) -> int:
        """
        The number of feature channels (the first volume dimension size).
        """
        return self.tensor.shape[0]

    @property
    def device(self) -> torch.device:
        """
        Device of the volume tensor.
        """
        return self.tensor.device

    @property
    def dtype(self) -> torch.dtype:
        """
        Datatype of the volume tensor.
        """
        return self.tensor.dtype

    def new(self,
        tensor: torch.Tensor,
        geometry: vx.AcquisitionGeometry = None) -> Volume:
        """
        Construct a new volume instance with the provided features tensor, while
        preserving any unchanged properties of the original volume.

        Args:
            tensor (Tensor): The new image tensor replacement.
            geometry (AcquisitionGeometry, optional): The new geometry. If None,
                the current geometry will be propagated.
        """
        geometry = self.geometry if geometry is None else geometry
        return self.__class__(tensor, geometry)

    def save(self, filename: os.PathLike, fmt: str = None) -> None:
        """
        Save the volume to a file.

        Args:
            filename (PathLike): The path to the file to save.
            fmt (str, optional): The format of the file. If None, the format is
                determined by the file extension.
        """
        vx.save_volume(self, filename, fmt=fmt)

    # -------------------------------------------------------------------------
    #  numerical and tensor operations for volume data manipulation
    # -------------------------------------------------------------------------

    def apply(self, func: callable) -> Volume:
        """
        Apply a function to the volume tensor and return a new instance.

        Args:
            func (callable): The function to apply.

        Returns:
            Volume: A new volume instance.
        """
        return self.new(func(self.tensor))

    def detach(self) -> Volume:
        """
        Detach the volume tensor from the current computational graph.

        Returns:
            Volume: A new volume instance with the detached tensor.
        """
        return self.new(self.tensor.detach(), self.geometry.detach())

    def to(self, device: torch.Device) -> Volume:
        """
        Move the volume tensor to a device.

        Args:
            device (Device): The target device.

        Returns:
            Volume: A new volume instance with the tensor on the target device.
        """
        if device is None:
            return self
        return self.new(self.tensor.to(device), self.geometry.to(device))

    def cuda(self) -> Volume:
        """
        Move the volume tensor to the GPU.

        Returns:
            Volume: A new volume instance with the tensor on the GPU.
        """
        return self.new(self.tensor.cuda(), self.geometry.cuda())

    def cpu(self) -> Volume:
        """
        Move the volume tensor to the CPU.

        Returns:
            Volume: A new volume instance with the tensor on the CPU.
        """
        return self.new(self.tensor.cpu(), self.geometry.cpu())

    def type(self, dtype: torch.dtype) -> Volume:
        """
        Convert the volume tensor to a specified data type.

        Args:
            dtype (torch.dtype): The target data type.

        Returns:
            Volume: A new volume instance.
        """
        if self.tensor.dtype == dtype:
            return self
        return self.new(self.tensor.type(dtype))

    def float(self) -> Volume:
        """
        Convert the volume tensor to float data type.

        Returns:
            Volume: A new float volume instance.
        """
        return self.new(self.tensor.float())

    def half(self) -> Volume:
        """
        Convert the volume tensor to half-precision float data type.

        Returns:
            Volume: A new half-precision float volume instance.
        """
        return self.new(self.tensor.half())

    def int(self) -> Volume:
        """
        Convert the volume tensor to integer data type.

        Returns:
            Volume: A new integer volume instance.
        """
        return self.new(self.tensor.int())

    def bool(self) -> Volume:
        """
        Convert the volume tensor to boolean data type.

        Returns:
            Volume: A new boolean volume instance.
        """
        return self.new(self.tensor.bool())

    def max(self, dim: int = None) -> Volume | torch.Tensor:
        """
        Get the maximum value in the volume tensor.

        Args:
            dim (int, optional): The dimension or dimensions to
                reduce. If None, all dimensions are reduced. If
                the dimension is 0 (channel axis), a single-channel
                volume is returned.

        Returns:
            Tensor or Volume: The maximum value(s) or volume.
        """
        reduced = self.tensor.amax(dim=dim)
        return self.new(reduced) if dim == 0 else reduced

    def min(self, dim: int = None) -> Volume | torch.Tensor:
        """
        Get the minimum value in the volume features.

        Args:
            dim (int, optional): The dimension or dimensions to
                reduce. If None, all dimensions are reduced. If
                the dimension is 0 (channel axis), a single-channel
                volume is returned.

        Returns:
            Tensor or Volume: The mininum value(s) or volume.
        """
        reduced = self.tensor.amin(dim=dim)
        return self.new(reduced) if dim == 0 else reduced

    def sum(self, dim: int = None) -> Volume | torch.Tensor:
        """
        Compute the sum of all voxels.

        Args:
            dim (int, optional): The dimension or dimensions to
                reduce. If None, all dimensions are reduced. If
                the dimension is 0 (channel axis), a single-channel
                volume is returned.

        Returns:
            Tensor or Volume: The summed value(s) or volume.
        """
        reduced = self.tensor.sum(dim=dim)
        return self.new(reduced) if dim == 0 else reduced

    def mean(self, dim: int = None) -> Volume | torch.Tensor:
        """
        Compute the mean of all voxels.

        Args:
            dim (int, optional): The dimension or dimensions to
                reduce. If None, all dimensions are reduced. If
                the dimension is 0 (channel axis), a single-channel
                volume is returned.

        Returns:
            Tensor or Volume: The mean value(s) or volume.
        """
        reduced = self.tensor.mean(dim=dim)
        return self.new(reduced) if dim == 0 else reduced

    def floor(self) -> Volume:
        """
        Apply the floor operation to the volume features.

        Returns:
            Volume: A new floored volume instance.
        """
        return self.new(self.tensor.floor())

    def ceil(self) -> Volume:
        """
        Apply the ceil operation to the volume features.

        Returns:
            Volume: A new ceiled volume instance.
        """
        return self.new(self.tensor.ceil())

    def abs(self) -> Volume:
        """
        Compute absolute values of the volume features.

        Returns:
            Volume: A new volume instance.
        """
        return self.new(self.tensor.abs())

    def exp(self) -> Volume:
        """
        Compute exponential of the elements in the volume features.

        Returns:
            Volume: A new exponentiated volume instance.
        """
        return self.new(self.tensor.exp())

    def log(self) -> Volume:
        """
        Compute the natural logarithm of the volume features.

        Returns:
            Volume: A new log-transformed volume instance.
        """
        return self.new(self.tensor.log())

    def sqrt(self) -> Volume:
        """
        Compute the square root of the volume features.

        Returns:
            Volume: A new square-rooted volume instance.
        """
        return self.new(self.tensor.sqrt())

    def square(self) -> Volume:
        """
        Compute the square of the volume features.

        Returns:
            Volume: A new squared volume instance.
        """
        return self.new(self.tensor.square())

    def pow(self, exponent: float) -> Volume:
        """
        Compute the power of the volume features.

        Args:
            exponent (float): The exponent value.

        Returns:
            Volume: A new powered volume instance.
        """
        return self.new(self.tensor.pow(exponent))

    def isnan(self) -> Volume:
        """
        Compute a mask of NaN values in the volume.

        Returns:
            Volume: A new volume mask instance.
        """
        return self.new(self.tensor.isnan())

    def clamp(self,
        min: float = None,
        max: float = None,
        inplace: bool = False) -> Volume:
        """
        Clamp the values in the volume tensor.

        Args:
            min (float, optional): Minimum value to clamp to.
            max (float, optional): Maximum value to clamp to.
            inplace (bool): Whether to perform the operation in-place.

        Returns:
            Volume: A new (if not in-place) clamped volume instance.
        """
        if inplace:
            return self.new(self.tensor.clamp_(min=min, max=max))
        else:
            return self.new(self.tensor.clamp(min=min, max=max))

    def maximum(self, other: Volume) -> Volume:
        """
        Computes the element-wise maximum between two volumes.

        Args:
            other (Volume): The input volume to compare against.

        Returns:
            Volume: A maximized volume instance.
        """
        return self.new(self.tensor.maximum(other.tensor))

    def minimum(self, other: Volume) -> Volume:
        """
        Computes the element-wise minimum between two volumes.

        Args:
            other (Volume): The input volume to compare against.

        Returns:
            Volume: A minimized volume instance.
        """
        return self.new(self.tensor.minimum(other.tensor))

    def all(self, dim: int = None) -> Volume | torch.Tensor:
        """
        Check if all elements in the volume are True.

        Args:
            dim (int, optional): The dimension or dimensions to
                reduce. If None, all dimensions are reduced. If
                the dimension is 0 (channel axis), a single-channel
                volume is returned.

        Returns:
            Tensor or Volume: The all-True value(s) or volume.
        """
        kwargs = {} if dim is None else {'dim': dim}
        reduced = self.tensor.all(**kwargs)
        return self.new(reduced) if dim == 0 else reduced

    def any(self, dim: int = None) -> Volume | torch.Tensor:
        """
        Check if any elements in the volume are True.

        Args:
            dim (int, optional): The dimension or dimensions to
                reduce. If None, all dimensions are reduced. If
                the dimension is 0 (channel axis), a single-channel
                volume is returned.

        Returns:
            Tensor or Volume: The any-True value(s) or volume.
        """
        kwargs = {} if dim is None else {'dim': dim}
        reduced = self.tensor.any(**kwargs)
        return self.new(reduced) if dim == 0 else reduced

    def zeros_like(self,
        channels: int = None,
        dtype: torch.dtype = None) -> Volume:
        """
        Create a volume of zeros with the same geometry and
        device as the current instance.

        Args:
            channels (int, optional): Number of channels in the new volume.
                If None, will default to the existing number.
            dtype (torch.dtype, optional): Target data type.

        Returns:
            Volume: A new volume instance filled with zeros.
        """
        if channels is None:
            channels = self.num_channels
        return self.geometry.zeros_like(channels, dtype=dtype)

    def ones_like(self,
        channels: int = None,
        dtype: torch.dtype = None) -> Volume:
        """
        Create a volume of ones with the same geometry and
        device as the current instance.

        Args:
            channels (int, optional): Number of channels in the new volume.
                If None, will default to the existing number.
            dtype (torch.dtype, optional): Target data type.

        Returns:
            Volume: A new volume instance filled with ones.
        """
        if channels is None:
            channels = self.num_channels
        return self.geometry.ones_like(channels, dtype=dtype)

    def full_like(self,
        fill: float,
        channels: int = None,
        dtype: torch.dtype = None) -> Volume:
        """
        Create a volume filled with a specific value and with the same
        geometry and device as the current instance.

        Args:
            fill (float): The fill value.
            channels (int, optional): Number of channels in the new volume.
                If None, will default to the existing number.
            dtype (torch.dtype, optional): Target data type.

        Returns:
            Volume: A new filled volume instance.
        """
        if channels is None:
            channels = self.num_channels
        return self.geometry.full_like(fill, channels, dtype=dtype)

    def rand_like(self,
        channels: int = None,
        dtype: torch.dtype = None) -> Volume:
        """
        Create a volume of random values with the same geometry and
        device as the current instance. Values are sampled from a uniform
        distribution on the interval [0, 1).

        Args:
            channels (int, optional): Number of channels in the new volume.
                If None, will default to the existing number.
            dtype (torch.dtype, optional): Target data type.

        Returns:
            Volume: A new random volume instance.
        """
        if channels is None:
            channels = self.num_channels
        return self.geometry.rand_like(channels, dtype=dtype)

    def randn_like(self,
        channels: int = None,
        dtype: torch.dtype = None) -> Volume:
        """
        Create a volume of random values with the same geometry and
        device as the current instance. Values are sampled from a normal
        distribution with mean 0 and variance 1

        Args:
            channels (int, optional): Number of channels in the new volume.
                If None, will default to the existing number.
            dtype (torch.dtype, optional): Target data type.

        Returns:
            Volume: A new random volume instance.
        """
        if channels is None:
            channels = self.num_channels
        return self.geometry.randn_like(channels, dtype=dtype)

    def isin(self, elements: torch.Tensor) -> Volume:
        """
        Tests if each element of `elements` is in the volume.

        Args:
            elements (Tensor or Scalar): Values against which to test each voxel.

        Returns:
            Volume: A boolean volume that is True when a voxel value is
                in `elements` and False otherwise.
        """
        if isinstance(elements, (list, tuple)):
            elements = torch.tensor(elements, device=self.device)
        return self.new(torch.isin(self.tensor, elements))

    def unique(self, **kwargs) -> torch.Tensor:
        """
        Compute the unique elements of volume.

        Args:
            **kwargs: Additional arguments passed to the underlying
                call to `torch.unique()`.

        Returns:
            Tensor: The output list of unique scalar elements.
        """
        return self.tensor.unique(**kwargs)

    def quantile(self, q: torch.Tensor) -> torch.Tensor:
        """
        Compute the q-th quantile of the voxel data. 

        Args:
            q (float): A scalar quantile in the range [0, 1].
        
        Returns:
            Tensor: The quantile scalar value.
        """
        if q < 0 or q > 1:
            raise ValueError(f'quantile must be between 0 and 1, got {q}')
        if q == 0:
            return self.tensor.min()
        if q == 1:
            return self.tensor.max()
        flattened = self.tensor.flatten()
        if q > 0.5:
            k = int(flattened.numel() * (1.0 - q)) + 1
            return flattened.topk(k, largest=True, sorted=False).values.min()
        else:
            k = int(flattened.numel() * q) + 1
            return flattened.topk(k, largest=False, sorted=False).values.max()

    def softmax(self, dim: int = 0) -> Volume | torch.Tensor:
        """
        Get the maximum index in the volume tensor.

        Args:
            dim (int, optional): The dimension or dimensions to
                reduce. If the dimension is 0 (channel axis), a
                single-channel volume is returned.

        Returns:
            Tensor or Volume: Softmaxed probabilities.
        """
        reduced = self.tensor.softmax(dim=dim)
        return self.new(reduced) if dim == 0 else reduced

    def argmax(self, dim: int = None) -> Volume | torch.Tensor:
        """
        Get the maximum index in the volume tensor.

        Args:
            dim (int, optional): The dimension or dimensions to
                reduce. If None, all dimensions are reduced. If
                the dimension is 0 (channel axis), a single-channel
                volume is returned.

        Returns:
            Tensor or Volume: The maximum indices or volume.
        """
        reduced = self.tensor.argmax(dim=dim)
        return self.new(reduced) if dim == 0 else reduced

    def onehot(self, num_classes: int = -1) -> vx.Volume:
        """
        One hot encode a label volume.

        Args:
            num_classes (int, optional): Total number of classes. If set to -1, the
                number of classes will be inferred as one greater than the largest
                class value in the input volume. 

        Returns:
            Tensor or Volume: The maximum indices or volume.
        """
        assert self.num_channels == 1, f'cannot one hot volume with {self.num_channels} channels'
        assert not torch.is_floating_point(self.tensor), f'one hot requires volume of type int, got {self.dtype}'
        tensor = torch.nn.functional.one_hot(self.tensor.squeeze(0).long(), num_classes=num_classes)
        return self.new(tensor.movedim(-1, 0))

    # -------------------------------------------------------------------------
    # indexing / operator overloads for tensor-style voxel data manipulation
    # -------------------------------------------------------------------------

    # assignment

    def __getitem__(self, indexing) -> torch.Tensor | Volume:
        # a regular boolean tensor-based indexing should be treated the
        # same as it would for a normal tensor
        if isinstance(indexing, torch.Tensor):
            return self.tensor[indexing]
        # the same goes for boolean volume indexing (in which case we'll
        # just use the underlying tensor)
        elif isinstance(indexing, Volume):
            indexing = indexing.tensor
            # if we get a one-channel boolean mask for the indexing,
            # we should auto-broadcast it to match the target channels
            if indexing.shape[0] == 1 and self.num_channels > 1:
                indexing = indexing.expand(self.num_channels, -1, -1, -1)
            return self.tensor[indexing]
        elif isinstance(indexing, list):
            # a list of indices should be treated as a list of channel reshuffling indices
            if not all(isinstance(i, int) for i in indexing):
                raise ValueError('channel list indexing must be a list of integers')
            return self.new(self.tensor[indexing])
        # in all circumstances (ex: slicing tuple or bounding box), call
        # the crop function which actually returns a new volume
        return self.crop(indexing)

    def __setitem__(self, indexing, value) -> None:
        self.tensor[_cast_volume_as_tensor(indexing)] = _cast_volume_as_tensor(value)

    # comparison operators

    def __eq__(self, other) -> Volume:
        return self.new(self.tensor == _cast_volume_as_tensor(other))
    
    def __ne__(self, other) -> Volume:
        return self.new(self.tensor != _cast_volume_as_tensor(other))

    def __lt__(self, other) -> Volume:
        return self.new(self.tensor < _cast_volume_as_tensor(other))

    def __le__(self, other) -> Volume:
        return self.new(self.tensor <= _cast_volume_as_tensor(other))

    def __gt__(self, other) -> Volume:
        return self.new(self.tensor > _cast_volume_as_tensor(other))

    def __ge__(self, other) -> Volume:
        return self.new(self.tensor >= _cast_volume_as_tensor(other))

    # unary operators

    def __pos__(self) -> Volume:
        return self.new(+self.tensor)

    def __neg__(self) -> Volume:
        return self.new(-self.tensor)

    # binary operators

    def __and__(self, other) -> Volume:
        return self.new(self.tensor & _cast_volume_as_tensor(other))

    def __or__(self, other) -> Volume:
        return self.new(self.tensor | _cast_volume_as_tensor(other))

    def __xor__(self, other) -> Volume:
        return self.new(self.tensor ^ _cast_volume_as_tensor(other))

    def __add__(self, other) -> Volume:
        return self.new(self.tensor + _cast_volume_as_tensor(other))

    def __radd__(self, other) -> Volume:
        return self.new(_cast_volume_as_tensor(other) + self.tensor)

    def __sub__(self, other) -> Volume:
        return self.new(self.tensor - _cast_volume_as_tensor(other))

    def __rsub__(self, other) -> Volume:
        return self.new(_cast_volume_as_tensor(other) - self.tensor)

    def __mul__(self, other) -> Volume:
        return self.new(self.tensor * _cast_volume_as_tensor(other))

    def __rmul__(self, other) -> Volume:
        return self.new(_cast_volume_as_tensor(other) * self.tensor)

    def __truediv__(self, other) -> Volume:
        return self.new(self.tensor / _cast_volume_as_tensor(other))

    def __rtruediv__(self, other) -> Volume:
        return self.new(_cast_volume_as_tensor(other) / self.tensor)

    def __pow__(self, other) -> Volume:
        return self.new(self.tensor ** _cast_volume_as_tensor(other))

    # assignment operators

    def __iadd__(self, other) -> None:
        self._tensor += _cast_volume_as_tensor(other)
        return self

    def __isub__(self, other) -> None:
        self._tensor -= _cast_volume_as_tensor(other)
        return self

    def __imul__(self, other) -> None:
        self._tensor *= _cast_volume_as_tensor(other)
        return self

    def __itruediv__(self, other) -> None:
        self._tensor /= _cast_volume_as_tensor(other)
        return self

    # -------------------------------------------------------------------------
    # methods for manipulating spatial geometry and computing coordinates
    # -------------------------------------------------------------------------

    def sample(self,
        points: torch.Tensor | vx.Mesh,
        space: vx.Space,
        mode: str = 'linear',
        padding_mode: str = 'zeros') -> torch.Tensor:
        """
        Sample volume features at a set of points.

        Args:
            points (Tensor | Mesh): A set of points in world or voxel coordinates with
                shape $(N, 3)$. If the input is a mesh, the vertex positions are used.
            space (Space): The coordinate space of the input points or mesh.
            mode (str, optional): The sampling mode, either 'linear' or 'nearest'.
            padding_mode (str, optional): Padding mode for outside grid values.

        Returns:
            Tensor: The sampled features, with shape $(N, C)$.
        """
        if isinstance(points, vx.Mesh):
            points = points.vertices

        # convert to local coordinate space
        if vx.Space(space) == 'world':
            points = self.geometry.inverse().transform(points)
        points = self.geometry.voxel_to_local_coordinates(points)

        # sample the channels
        sampled = torch.nn.functional.grid_sample(
            self.tensor.float().unsqueeze(0),
            points.view(1, len(points), 1, 1, 3),
            align_corners=False,
            mode=('bilinear' if mode == 'linear' else 'nearest'),
            padding_mode=padding_mode)
        
        # if nearest neighbor sampling, convert back to original dtype
        if mode == 'nearest':
            sampled = sampled.type(self.dtype)

        # remove batch and spatial dimensions
        return sampled.squeeze(dim=(0, 3, 4)).swapaxes(0, 1)

    def tesselate(self, threshold: float = 0.5, space: vx.Space = 'world') -> vx.Mesh:
        """
        Tesselate a mesh around connected voxel components.
        This is not differentiable.

        Args:
            threshold (float, optional): Scalar threshold that determines
                whether a voxel is inside or outside the mesh boundary.
            space (Space, optional): The coordinate space of mesh vertices. Default
                is the world coordinate space.

        Returns:
            Mesh: Tesselated mesh.
        """
        try:
            from pytorch3d.ops.marching_cubes import marching_cubes
        except ImportError as exc:
            raise ImportError('mesh tesselation requires that the '
                              'pytorch3d package is installed') from exc

        # 
        padded = self.detach().pad(1, 'voxel')
        vertices, faces = marching_cubes(padded.tensor.float(), threshold,
                                         return_local_coords=False)
        if len(vertices[0]) == 0:
            raise ValueError('empty volume - could not tesselate')
        mesh = vx.Mesh(vertices[0].flip(-1) - 1, faces[0])

        # 
        if vx.Space(space) == 'world':
            mesh = mesh.transform(self.geometry)
        return mesh

    def bounds(self,
        nonzero: bool = False,
        margin: float | torch.Tensor = None,
        space: vx.Space = 'world') -> vx.Mesh:
        """
        Compute a box mesh enclosing the bounds of the volume grid or the non-zero
        voxels in the image.

        Args:
            nonzero (bool): If True, compute the bounds around all non-zero voxels,
                otherwise use the extent of the image grid.
            margin (float or Tensor, optional): Margin to expand the cropping boundary.
                Can be a positive or negative delta.
            space (Space, optional): Space of the margin values, either 'voxel' or 'world'.

        Returns:
            Mesh: Bounding box mesh in world-space coordinates.
        """
        if nonzero:
            # compute the bounding box around all nonzero voxels
            tensor = self.tensor if self.num_channels > 1 else self.tensor.sum(0)
            nonzero = tensor.view(self.baseshape).nonzero()
            if nonzero.shape[0] == 0:
                raise ValueError('cannot compute nonzero bounds on an empty volume')
            min_point = nonzero.amin(dim=0).float().cpu()
            max_point = nonzero.amax(dim=0).float().cpu()
        else:
            # just use the bounds of the volume extent
            min_point = torch.zeros(3)
            max_point = torch.tensor(self.baseshape).float() - 1

        # expand (or shrink) margin around border
        if margin is not None:
            margin = self.geometry.conform_units(margin, space, 'voxel', 2).cpu()
            min_point -= margin[:, 0]
            max_point += margin[:, 1]

        # build the world-space bounding box mesh
        mesh = vx.mesh.construct_box_mesh(min_point, max_point)
        return mesh.transform(self.geometry)

    def centroids(self, space: vx.Space) -> torch.Tensor:
        """
        Compute the centroids (centers of mass) for each volume channel.
        All negative values are clamped to zero before computing the centroids.

        Args:
            space (Space): The coordinate space of computed centroids.

        Returns:
            Tensor: Per-channel coordinates of shape (C, 3).
        """
        clamped_tensor = self.tensor.clamp(min=0).float()

        # compute the centroids with a differentiable coordinate weighting
        coord = lambda a: (a * torch.arange(a.shape[-1], device=self.device)).sum(-1) / (a.sum(-1) + 1e-6)
        z_mean = clamped_tensor.mean(-1)
        x = coord(z_mean.mean(-1))
        y = coord(z_mean.mean(-2))
        z = coord(clamped_tensor.mean(-2).mean(-2))
        centroids = torch.stack([x, y, z], dim=-1)

        # transform to world-space if necessary
        if vx.Space(space) == 'world':
            centroids = self.geometry.transform(centroids)
        return centroids

    def slice(self,
        point: int | torch.Tensor,
        direction: int | torch.Tensor,
        space: vx.Space) -> Volume:
        """
        Extract a slice from the volume. Note this will still return a volume,
        but with a slice dimension reduced to 1.

        Args:
            point (int or Tensor): A point of the slice plane. If a tensor,
                it should represent a 3D point coordinate. If an int, it should be
                the index of the slice in the specified direction. Note that this requires
                the slice direction axis to be specified as an int as well.
            direction (int or Tensor): The direction of the slice plane. If a tensor,
                it should represent a 3D vector direction. If an int, it should be
                the index of the slice in the specified direction.
            space (Space): The coordinate space of the slice point and direction.

        Returns:
            Volume: The sliced volume instance.
        """
        if vx.Space(space) == 'world':
            raise NotImplementedError('slicing in world space is not yet supported')
        if isinstance(point, torch.Tensor):
            raise NotImplementedError('slicing with a 3d plane point is not yet supported')
        if isinstance(direction, torch.Tensor):
            raise NotImplementedError('slicing with a direction vector is not yet supported')

        if direction < 0 or direction > 2:
            raise ValueError(f'slice direction must be between 0 and 2, got {direction}')
        if point < 0 or point >= self.baseshape[direction]:
            raise ValueError(f'slice index {point} out of bounds for shape {self.baseshape}')

        # create a cropping tuple to extract the slice
        cropping = [slice(None) for _ in range(4)]
        cropping[direction + 1] = slice(point, point + 1)
        return self[tuple(cropping)]

    def crop(self,
        cropping: tuple | vx.Mesh,
        margin: float | torch.Tensor = None,
        space: vx.Space = 'world') -> Volume:
        """
        Crop the volume to some bounding, either defined by a voxel slicing
        tuple or a bounding box mesh.

        Args:
            cropping (tuple or Mesh): Cropping defined by either a tuple of slices
                or a bounding box mesh.
            margin (float or Tensor, optional): Margin to expand the cropping boundary.
                Can be a positive or negative delta. The boundary will be clipped if it
                extends beyond the shape of the volume.
            space (Space): The coordinate space of the margin values, either
                'voxel' or 'world'.

        Returns:
            Volume: The cropped volume instance.
        """

        # transform to voxel units
        if margin is not None:
            margin = self.geometry.conform_units(margin, space, 'voxel').cpu().round().int()

        if isinstance(cropping, vx.Mesh):
            # if we get a mesh as input, assume its a bounding box, but really
            # any set of mesh points could work here
            world2voxel = self.geometry.inverse()
            points = world2voxel.transform(cropping.vertices.detach())
            minc = points.amin(0).cpu().ceil().int()
            maxc = points.amax(0).cpu().floor().int()

            # extend the boundary
            if margin is not None:
                minc -= margin
                maxc += margin

            # make sure the coordinates are clamped within the volume extent
            minc = minc.clamp(min=0)
            maxc = maxc.clamp(max=torch.tensor(self.baseshape))
            stride = None

            # convert coordinate bounds to a 4D slicing tuple
            slicing = (slice(None), *vx.slicing.coordinates_to_slicing(minc, maxc))

        elif isinstance(cropping, (tuple, int, slice, type(...))):

            # conform single indexing items to a tuple format
            if not isinstance(cropping, tuple):
                cropping = (cropping,)

            # if we get a tuple assume its a tuple of slices
            slicing = vx.slicing.expand_slicing(cropping, 4)

            # do not allow cropping to remove a spatial dimension
            if any(isinstance(s, int) for s in slicing[1:]):
                raise ValueError('cannot remove a spatial dimension when cropping a volume')

            # extend the boundary
            minc, maxc, stride = vx.slicing.slicing_to_coordinates(cropping[1:], self.baseshape)
            if margin is not None:
                minc = (minc - margin).clamp(min=0)
                maxc = (maxc + margin).clamp(max=torch.tensor(self.baseshape))
                slicing = (slicing[0], *vx.slicing.coordinates_to_slicing(minc, maxc, stride))
        else:
            raise ValueError(f'unknown cropping item: {type(cropping)}')

        # update the geometry based on any inferred voxel shift or scale
        geometry = self.geometry
        if any(minc != 0):
            geometry = geometry.shift(minc, space='voxel')
        if stride is not None:
            geometry = geometry.scale(stride, space='voxel')

        # apply the cropping
        cropped_tensor = self.tensor[slicing]
        cropped_geometry = vx.AcquisitionGeometry(baseshape=cropped_tensor.shape[-3:],
                                                  matrix=geometry.tensor,
                                                  slice_direction=geometry._explicit_slice_direction)
        return self.new(cropped_tensor, cropped_geometry)

    def crop_to_nonzero(self, margin: float  | torch.Tensor = None) -> Volume:
        """
        Crop the volume to the bounding box around nonzero voxels.

        Args:
            margin (float or Tensor, optional): Margin (in world units) to expand
                the cropping boundary. Can be a positive or negative delta. The
                boundary will be clipped if it extends beyond the shape of the volume.

        Returns:
            Volume: The cropped volume instance.
        """
        nonzero = self.tensor.nonzero()
        # note: we're using nonzero() directly here instead of calling self.bounds() to avoid
        # the unnecessary transformation into world space then back again)
        minc = nonzero.amin(0)
        maxc = nonzero.amax(0)
        return self.crop(vx.slicing.coordinates_to_slicing(minc, maxc), margin=margin)

    def reorient(self, orientation: vx.Orientation) -> Volume:
        """
        Transform the volume to a new orientation. This is faster than
        achieving the same result through resampling.

        Args:
            orientation (Orientation): The target orientation.
        
        Returns:
            Volume: The reoriented volume instance.
        """
        source = self.geometry.orientation
        target = vx.cast_orientation(orientation)

        perm = source.dims.argsort()[target.dims]
        tensor = self.tensor
        if (perm != torch.tensor((0, 1, 2))).any():
            tensor = tensor.permute(0, *(perm + 1))

        flip = source.flip[perm] * target.flip
        indices = (flip < 0).argwhere()
        if len(indices) > 0:
            tensor = tensor.flip(*(indices + 1))

        return self.new(tensor, self.geometry.reorient(orientation))

    def resample_like(self,
        target: Volume | vx.AcquisitionGeometry,
        mode: str = 'linear',
        padding_mode: str = 'zeros',
        antialias: bool = False) -> Volume:
        """
        Resample the volume features to match the geometry of a target volume.

        Args:
            target (Volume | AcquisitionGeometry): Target acquisition geometry.
            mode (str, optional): Interpolation mode.
            padding_mode (str, optional): Padding mode for outside grid values.
            antialias (bool, optional): If True, will apply a Gaussian filter
                before resampling to avoid aliasing artifacts.

        Returns:
            Volume: Resampled volume instance.
        """
        if isinstance(target, Volume):
            target = target.geometry

        # check if the matrices are similar because we might be able to avoid any
        # actual resampling if that's the case. first, we check the rotation and scale
        if torch.allclose(self.geometry.tensor[:, :3], target.tensor[:, :3], atol=1e-4, rtol=0):

            # then check if the source and target have matching baseshapes and matrices, because
            # in that case we don't have to modify anything at all
            if target.baseshape == self.baseshape and \
               torch.allclose(self.geometry.tensor[:3, -1], target.tensor[:3, -1], atol=1e-4, rtol=0):
                return self.new(self.tensor, target)

            # otherwise, it's possible the difference between image spaces is only a voxel-shift,
            # in which case we can just crop and/or pad -- much faster than resampling.
            # we need to check if the voxel-space translations are all integers
            delta = (self.geometry.inverse() @ target).transform(torch.zeros(3))
            delta_rounded = delta.round()
            if torch.allclose(delta, delta_rounded, atol=1e-4, rtol=0):

                # these are the relative shifts in voxels at the lower (origin) and upper corners
                lower = delta_rounded.int().cpu()
                upper = lower + torch.tensor(target.baseshape) - torch.tensor(self.baseshape)

                # apply any necessary cropping to the tensor
                minc = lower.clamp(min=0)
                maxc = upper.clamp(max=0) + torch.tensor(self.baseshape)
                slicing = (slice(None), *[slice(a, b) for a, b in zip(minc, maxc)])
                resampled = self.tensor[slicing]

                # apply any necessary padding to the tensor
                a = lower.clamp(max=0).abs()
                b = upper.clamp(min=0)
                padding = torch.stack((b, a), dim=1).flatten()
                if (padding != 0).any():
                    mode = dict(zeros='constant', reflection='reflect', border='replicate').get(padding_mode)
                    if mode is None:
                        raise ValueError(f'no padding mode equivolent for {padding_mode}')
                    reverse = list(reversed([int(d) for d in padding]))
                    resampled = torch.nn.functional.pad(resampled, reverse, mode=mode)

                return self.new(resampled, target)
    
        # if we got here, it means have to resort to doing a grid interpolation, so first
        # build the coordinate grid for the target image
        transform = self.geometry.inverse() @ target

        if antialias:
            # if antialiasing is enabled, we'll need to do some extra work. the goal is to smooth
            # intensities along a downsampling direction. since there is no guarantee that the
            # target geometry dimensions are aligned with the source grid (e.g. consider a rotation
            # resampling), we can't just smooth the source image and sample it using the target grid.
            # instead, we need to compute an intermediate grid that is aligned with the target so that
            # we can apply smoothing kernels in each downsampling directions of the target geometry.
            # the intermediate grid is a multiple of the target grid (determined by the downsample factor)
            # and we can use strided convolutions to get the final resampled image efficiently.

            if mode != 'linear':
                raise ValueError('antialiasing only supported with linear interpolation')

            inter_space = vx.AcquisitionGeometry(target.baseshape, transform)
            down_factor = inter_space.spacing.clamp(1).floor().int()

            # blur with a sigma of 1/3 of the downsample factor
            sigma = (down_factor > 1) * (down_factor.float() / 3)

            # compute the padding required for the Gaussian kernel (hardcoded the truncate value)
            # along with the updated grid transform
            truncate = 2
            padding = (truncate * sigma + 0.5).int()
            intermediate_baseshape = [int(s * f) + p * 2 for s, f, p in zip(target.baseshape, down_factor, padding)]
            transform = inter_space.scale(1 / down_factor, space='voxel').shift(-padding, space='voxel')
        else:
            intermediate_baseshape = target.baseshape

        grid = volume_grid(intermediate_baseshape, transform=transform,
                           localshape=self.baseshape, device=self.device)

        resampled = torch.nn.functional.grid_sample(
                        input=self.tensor.float().unsqueeze(0),
                        grid=grid.unsqueeze(0),
                        mode=('bilinear' if mode == 'linear' else mode),
                        padding_mode=padding_mode,
                        align_corners=False).squeeze(0)

        if antialias:
            resampled = vx.filters.gaussian_blur(resampled, sigma, stride=tuple(down_factor),
                                                 padding='valid', truncate=truncate)

        # probably ideal to keep the data type consistent when using nearest neighbor sampling
        if mode == 'nearest':
            resampled = resampled.type(self.dtype)

        return self.new(resampled, target)

    def resample(self,
        spacing: float | torch.Tensor = None,
        in_plane_spacing: float | torch.Tensor = None,
        slice_spacing: float | torch.Tensor = None,
        mode: str = 'linear',
        padding_mode: str = 'zeros',
        antialias: bool = False) -> Volume:
        """
        Resample voxel features to a new voxel grid spacing.

        Args:
            spacing (float |Tensor): Target voxel spacing. An isotropic target
                is assumed if a scalar is provided.
            in_plane_spacing (float | Tensor): Target in-plane voxel spacing. Mutually
                exclusive with the `spacing` argument.
            slice_spacing (float | Tensor): Target slice spacing. Mutually exclusive
                except with the `spacing` argument.
            mode (str, optional): Interpolation mode.
            padding_mode (str, optional): Padding mode for outside grid values.
            antialias (bool, optional): If True, will apply a Gaussian filter
                before resampling to avoid aliasing artifacts.

        Returns:
            Volume: Volume resampled to the target voxel spacing.
        """
        target = self.geometry.resample(spacing=spacing, in_plane_spacing=in_plane_spacing,
                                        slice_spacing=slice_spacing)
        return self.resample_like(target, mode=mode, padding_mode=padding_mode, antialias=antialias)

    def reshape(self, baseshape: torch.Size) -> Volume:
        """
        Modify the spatial extent of the volume, cropping or padding around the
        center image to fit a given **baseshape**.

        This method is symmetric in that performing a reverse reshape operation
        will always yield the original geometry.

        args:
            baseshape (torch.Size): Target spatial (3D) shape.

        returns:
            Volume: Reshaped volume instance.
        """
        return self.resample_like(self.geometry.reshape(baseshape), mode='nearest')

    def pad(self, delta: float | torch.Tensor, space: vx.Space) -> Volume:
        """
        Pad the spatial extent of the volume by a given delta. Note that
        a negative delta value will result in trimming (cropping).

        args:
            delta (float or Tensor, optional): Delta of specified units to
                pad (or crop) the volume by in each direction. Can be of size
                $(1,)$, $(3,)$, or $(3, 2)$.
            space (Space): The coordinate space of the delta values, either
                'voxel' or 'world'.

        returns:
            Volume: Reshaped volume instance.
        """
        return self.resample_like(self.geometry.pad(delta, space), mode='nearest')

    def trim(self, delta: float | torch.Tensor, space: vx.Space) -> Volume:
        """
        Trim the spatial extent of the volume by a given delta. This is
        equivalent to padding with negative delta values.

        args:
            delta (float or Tensor, optional): Delta of specified units to
                trim the volume by in each direction. Can be of size $(1,)$,
                $(3,)$, or $(3, 2)$.
            space (Space): The coordinate space of the delta values, either
                'voxel' or 'world'.

        returns:
            Volume: Reshaped volume instance.
        """
        return self.pad(-delta, space)

    def transform(self,
        transform: vx.AffineVolumeTransform | vx.AffineMatrix,
        resample: bool = False,
        negate: bool = False,
        mode: str = 'linear',
        padding_mode: str = 'zeros') -> Volume:
        """
        Apply a spatial transform to the volume. By default, this method will not
        resample the image data and instead transform the world geometry.

        Args:
            transform (AffineVolumeTransform or AffineMatrix): Transform to apply. Assume
                a world-space transform if an AffineMatrix is provided.
            resample (bool, optional): If True, the volume will be transformed and
                resampled in voxel space, otherwise only the geometry will be updated.
            negate (bool, optional): If True, the inverse transform is applied to the
                geometry so that image features do not move in world space. This option
                can only be enabled when resampling is enabled.
            mode (str, optional): Interpolation mode if resampling.

        Returns:
            Volume: Transformed volume.
        """

        # if the transform is just a simple matrix, assume it's a world-space transform
        if not isinstance(transform, vx.AffineVolumeTransform):
            transform = vx.AffineVolumeTransform(transform, space='world', source=self, target=self)

        if not resample:
            # just apply the transform to the acquisition geometry
            if negate:
                raise ValueError('cannot negate transform when resampling is disabled')
            transform = transform.convert(space='world')
            return self.new(self.tensor, transform @ self.geometry)

        # if we're resampling, convert to a voxel-to-voxel transform
        target = transform.target
        inverted = transform.convert(space='voxel', source=self).inverse()

        # construct the transformed resampling grid
        grid = volume_grid(target.baseshape, transform=inverted,
                           localshape=self.baseshape, device=self.device)

        interpolated = torch.nn.functional.grid_sample(
                        self.tensor.unsqueeze(0).float(),
                        grid.unsqueeze(0),
                        mode=('bilinear' if mode == 'linear' else mode),
                        padding_mode=padding_mode,
                        align_corners=False).squeeze(0)

        if negate:
            # apply inverse transform to the geometry to cancel out world space changes
            target = inverted.convert(space='world') @ target

        return self.new(interpolated, target)

    def pool(self,
        scale: int = 2,
        mode: str = 'mean',
        spacing_ratio_thresh: float = None) -> Volume:
        """
        Pool the voxel data with a sliding window.

        By default, this will pool over all dimensions, but it can be conditionally
        disabled for the slice dimension based on the ratio of slice vs in-plane spacing, 
        i.e. the value of `geometry.spacing_ratio`. For example, if the slice spacing is
        `spacing_ratio_thresh` times greater than the in-plane spacing, the slice dimension
        will not be pooled. Mind that if the resulting pooled volume has a slice spacing
        less than the in-plane spacing, it will be resampled to an isotropic resolution.

        There is no analogous `unpool` method because there is complexity in determining
        the desired unpooling strategy. To return to the original geometry, instead use
        the `resample_like` method. If no reference geometry is available, just use
        the `reshape` method to upsample.

        Note that this implementation must mirror the pooling operation used by the geometry
        class. Any changes to the pooling operation in one class must be reflected in the other.

        Args:
            scale (int, optional): The size of the pooling window. Defaults to 2.
            mode (str, optional): Pooling mode - can be 'mean' or 'max'. Defaults to 'mean'.
            spacing_ratio_thresh (float, optional): Slice spacing ration that determines
                whether the slice dimension is pooled. This is disabled by default.

        Returns:
            Volume: Pooled volume.
        """

        # dowsample factors for each dimension
        factors = [1 if d == 1 else scale for d in self.baseshape]

        # check if we should pool the slice dimension based on the spacing ratio threshold
        spacing_ratio = self.geometry.spacing_ratio
        slice_dim_pooling = spacing_ratio_thresh is None or spacing_ratio < spacing_ratio_thresh

        if not slice_dim_pooling:
            factors[self.geometry.slice_direction] = 1

        # apply the appropriate pooling operation
        if mode == 'max':
            func = torch.nn.functional.max_pool3d
        elif mode == 'mean':
            func = torch.nn.functional.avg_pool3d
        else:
            raise ValueError(f'unknown pooling mode \'{mode}\'')

        pooled_tensor = func(self.tensor, factors, ceil_mode=True)

        # adjust the geometry based on the pooling factors
        shift = [0.5 * (f - 1) for f in factors]
        adjusted = self.geometry.shift(shift, space='voxel').scale(factors, space='voxel')
        pooled = self.new(pooled_tensor, vx.AcquisitionGeometry(pooled_tensor.shape[1:], adjusted))

        # if the slice dimension was not pooled and the resulting slice spacing
        # is less than the in-plane spacing, we need to resample the slice dimension
        if not slice_dim_pooling and spacing_ratio < scale:
            spacing = pooled.geometry.spacing.clone()
            spacing[self.geometry.slice_direction] = spacing[self.geometry.in_plane_directions].mean()
            pooled = pooled.resample(spacing, padding_mode='border')

        # sanity check
        if pooled.geometry.spacing_ratio < 0.99:
            raise ValueError('unexpected spacing ratio after pooling operation')

        return pooled

    # -------------------------------------------------------------------------
    # image filtering and statistical normalization
    # -------------------------------------------------------------------------

    def smooth(self, sigma: float | torch.Tensor, truncate: float = 2) -> Volume:
        """
        Apply Gaussian smoothing to the image features.

        Args:
            sigma (float | Tensor): Smoothing sigma in world space units.
            truncate (float, optional): The number of standard deviations to extend
            the kernel before truncating.

        Returns:
            Volume: Smoothed volume.
        """
        scaled = torch.as_tensor(sigma, device=self.device) / self.geometry.spacing.to(self.device)
        return self.new(vx.filters.gaussian_blur(self.tensor, scaled, truncate=truncate))

    def dilate(self, iterations: int = 1) -> Volume:
        """
        Dilate the volume.

        Args:
            iterations (int, optional): Number of dilation iterations.

        Returns:
            Volume: Dilated volume of the same data type.
        """
        return self.new(vx.filters.dilate(self.tensor, iterations))


def _cast_volume_as_tensor(other: object) -> object:
    """
    If provided a Volume, cast to a Tensor, otherwise return the input.
    """
    return other.tensor if isinstance(other, Volume) else other


def volume_grid(
    baseshape: torch.Size,
    transform: vx.AffineMatrix = None,
    localshape: torch.Size = None,
    device: torch.device = None) -> torch.Tensor:
    """
    Construct a grid of 3D voxel coordinates of the shape (W, H, D, 3).

    Args:
        baseshape (Size): Spatial (3D) shape of the volume grid.
        transform (AffineMatrix, optional): Grid voxel coordinate transform.
        localshape (Size, optional): If True, the grid is normalized
            between [1, -1] using the provided spatial shape and the
            coordinate order is swapped (for torch sampling methods).
        device (torch.device, optional): Device on which to
            allocate the grid data.

    Returns:
        Tensor: Grid volume tensor.
    """
    ranges = [torch.arange(s, dtype=torch.float32, device=device) for s in baseshape]
    grid = torch.stack(torch.meshgrid(*ranges, indexing='ij'), dim=-1)
    if transform:
        grid = transform.transform(grid)
    if localshape is not None:
        shape = torch.tensor(localshape).to(grid.device)
        grid = ((2 * grid + 1) / shape - 1).flip(-1)
    return grid


def stack(*vols):
    """
    Concatenates (stacks) multiple volumes channel-wise. Assumes
    volumes are in the same image space (with the same base shape).

    Args:
        *vols (sequence of Volumes): Volumes to merge.

    Returns:
        Volume: Single channel-stacked volume instance.
    """
    if len(vols) == 1 and not isinstance(vols[0], Volume):
        vols = vols[0]
    if len(vols) == 1:
        return vols[0]
    return vols[0].new(torch.cat([v.tensor for v in vols], dim=0))


def volumes_equal(
    a: Volume,
    b: Volume,
    vol_tol: float = 1e-6,
    geom_tol: float = 1e-6) -> bool:
    """
    Check if two volumes are equal within a given tolerance.

    Args:
        a, b (Volume): Volumes to compare.
        vol_tol (float, optional): Absolute tolerance for volume tensor comparison.
        geom_tol (float, optional): Absolute tolerance for geometry comparison.

    Returns:
        bool: True if the volumes are equal, False otherwise.
    """
    if not a.tensor.allclose(b.tensor, atol=vol_tol, rtol=0):
        return False
    if not a.geometry.tensor.allclose(b.geometry.tensor, atol=geom_tol, rtol=0):
        return False
    return True



"""
Reading and writing image volumes to various file formats.
"""

import os
import torch
import numpy as np
import voxel as vx

from .utility import IOProtocol


def load_volume(filename: os.PathLike, fmt: str = None) -> vx.Volume:
    """
    Load a volume from a file.

    Args:
        filename (PathLike): The path to the file to load.
        fmt (str, optional): The format of the file. If None, the format is
            determined by the file extension.

    Returns:
        Volume: The loaded volume.
    """
    vx.io.utility.check_file_readability(filename)

    if fmt is None:
        proto = vx.io.utility.find_protocol_by_extension(volume_io_protocols, filename)
        if proto is None:
            raise ValueError(f'cannot determine file format from extension for {filename}')
    else:
        proto = vx.io.protocol.find_protocol_by_name(volume_io_protocols, fmt)
        if proto is None:
            raise ValueError(f'unknown file format {fmt}')

    return proto().load(filename)


def save_volume(volume: vx.Volume, filename: os.PathLike, fmt: str = None) -> None:
    """
    Save a volume to a file.

    Args:
        volume (Volume): The volume to save.
        filename (PathLike): The path to the file to save.
        fmt (str, optional): The format of the file. If None, the format is
            determined by the file extension.
    """
    if fmt is None:
        proto = vx.io.utility.find_protocol_by_extension(volume_io_protocols, filename)
        if proto is None:
            raise ValueError(f'cannot determine file format from extension for {filename}')
    else:
        proto = vx.io.protocol.find_protocol_by_name(volume_io_protocols, fmt)
        if proto is None:
            raise ValueError(f'unknown file format {fmt}')
        filename = proto.enforce_extension(filename)

    proto().save(volume, filename)


class NiftiHeaderReference:
    """
    A reference to cache parameters of a nifti file header.
    This is passed around in the metadata of an acquisition
    geometry so that use it a reference (if needed) for resaving
    a volume without introducing any corruptions to the original
    file header.v
    """

    def __init__(self, nii) -> None:
        """
        Args:
            nii (Nifti1Image): The nifti image to cache.
        """
        self.qform_code = int(nii.header['qform_code'])
        self.sform_code = int(nii.header['sform_code'])
        self.xyzt_units = nii.header['xyzt_units']
        self.channel_spacing = nii.header['pixdim'][4]
        self.baseshape = tuple(nii.header['dim'][1:4])
        self.spacing = nii.header['pixdim'][1:4]
        self.affine = nii.affine


class NiftiArrayIO(IOProtocol):
    """
    Array IO protocol for nifti files.
    """
    name = 'nifti'
    extensions = ('.nii.gz', '.nii')

    def __init__(self) -> None:
        try:
            import nibabel as nib
        except ImportError:
            raise ImportError('the `nibabel` python package must be installed for nifti volume IO')
        self.nib = nib

    def load(self, filename: os.PathLike) -> vx.Volume:
        """
        Read array from a Nifti file.

        Args:
            filename (PathLike): The path to the Nifti file to read.

        Returns:
            Volume: The loaded volume.
        """
        nii = self.nib.load(filename)
        array = np.asanyarray(nii.dataobj)

        # not supported by torch
        if array.dtype == np.uint16:
            array = array.astype(np.int32)

        features = torch.from_numpy(array)
        if features.ndim == 4:
            features = features.moveaxis(-1, 0)

        # 
        spacing = torch.from_numpy(nii.header['pixdim'][1:4])
        affine = torch.from_numpy(nii.affine)
        volume = vx.Volume(features, affine)

        # 
        volume.geometry.reference['nii'] = NiftiHeaderReference(nii)

        # 
        if not torch.isclose(volume.geometry.spacing, spacing, atol=0.01, rtol=0.2).all():
            explicit_spacing = ', '.join([f'{s}:.2f' for s in spacing])
            affine_spacing = ', '.join([f'{s}:.2f' for s in volume.geometry.spacing])
            print('Warning: There is a substantial difference between the explicit voxel '
                  f'spacing in the nifti header ({explicit_spacing}) and the '
                  f'computed spacing from the scanner affine ({affine_spacing})')

        # 
        return volume

    def save(self, volume: vx.Volume, filename: os.PathLike) -> None:
        """
        Write volume to a Nifti file.

        Args:
            volume (Volume): The volume to save.
            filename (PathLike): The path to the Nifti file to write.
        """

        # 
        volume_array = volume.tensor.movedim(0, -1).detach().cpu().numpy()
        if volume_array.shape[-1] == 1:
            volume_array = np.squeeze(volume_array, -1)

        # convert to a valid output type (for now this is only bool but there are probably more)
        type_map = {np.dtype('bool'): np.uint8}
        dtype_id = next((i for dt, i in type_map.items() if np.issubdtype(volume_array.dtype, dt)), None)
        if dtype_id is not None:
            volume_array = volume_array.astype(dtype_id)

        # 
        nii = self.nib.Nifti1Image(volume_array, np.eye(4))

        # 
        spacing = volume.geometry.spacing.detach().cpu().numpy().astype(np.float64)
        affine = volume.geometry.tensor.detach().cpu().numpy().astype(np.float64)

        # 
        ref = volume.geometry.reference.get('nii')
        matches_original = ref is not None and \
            ref.baseshape == tuple(volume.baseshape) and \
            np.isclose(ref.affine, affine, rtol=0, atol=1e-3).all()

        # 
        if matches_original:
            affine = ref.affine
            spacing = ref.spacing

        # set spatial and temporal spacing
        nii.header['pixdim'][:] = 1
        nii.header['pixdim'][4] = 1 if not matches_original else ref.channel_spacing
        nii.header['pixdim'][1:4] = spacing

        # set units - fallback to mm and seconds
        default = np.asarray(2, dtype=np.uint8) | np.asarray(8, dtype=np.uint8)
        nii.header['xyzt_units'] = default if not matches_original else ref.xyzt_units

        # geometry-specific header data
        nii.set_sform(affine, 1 if not matches_original else ref.sform_code)
        nii.set_qform(affine, 1 if not matches_original else ref.qform_code)

        # write
        self.nib.save(nii, filename)


class MghArrayIO(IOProtocol):
    """
    Array IO protocol for mgh files.
    """
    name = 'mgh'
    extensions = ('.mgz', '.mgh')

    def __init__(self) -> None:
        try:
            import surfa as sf
        except ImportError:
            raise ImportError('the `surfa` python package must be installed for mgh volume IO')
        self.sf = sf

    def load(self, filename: os.PathLike) -> vx.Volume:
        """
        Read array from a MGH file.

        Args:
            filename (PathLike): The path to the MGH file to read.

        Returns:
            Volume: The loaded volume.
        """
        sv = self.sf.load_volume(filename)

        data = vx.io.numpy_to_tensor(sv.framed_data).movedim(-1, 0)
        matrix = vx.io.numpy_to_tensor(sv.geom.vox2world.matrix, copy=True)
        volume = vx.Volume(data, matrix)

        volume.geometry.reference['mgh'] = sv.geom

        return volume

    def save(self, volume: vx.Volume, filename: os.PathLike) -> None:
        """
        Write volume to a MGH file.

        Args:
            volume (Volume): The volume to save.
            filename (PathLike): The path to the MGH file to write.
        """
        volume_array = volume.tensor.movedim(0, -1).detach().cpu().numpy()
        affine = volume.geometry.tensor.detach().cpu().numpy()

        ref = volume.geometry.reference.get('mgh')
        matches_original = ref is not None and \
            tuple(ref.shape) == tuple(volume.baseshape) and \
            np.isclose(ref.vox2world.matrix, affine, rtol=0, atol=1e-3).all()

        geometry = ref if matches_original else self.sf.ImageGeometry(volume.baseshape, vox2world=affine)

        self.sf.Volume(volume_array, geometry=geometry).save(filename)


class PytorchVolumeIO(IOProtocol):
    """
    Array IO protocol for storing a simple volume in a pytorch file.
    The only data stored is the features tensor and the world affine.
    This is useful for fast data loading during training.
    """
    name = 'torch'
    extensions = ('.pth', '.pt')

    def load(self, filename: os.PathLike) -> vx.Volume:
        """
        Read array from a pytorch file.

        Args:
            filename (PathLike): The path to the pytorch file to read.

        Returns:
            Volume: The loaded volume.
        """
        items = torch.load(filename, weights_only=False)
        if 'v' not in items or 'm' not in items:
            raise RuntimeError(f'could not find `v` or `m` data keys in {filename}')
        return vx.Volume(items['v'], items['m'])

    def save(self, volume: vx.Volume, filename: os.PathLike) -> None:
        """
        Write volume to a pytorch file.

        Args:
            volume (Volume): The volume to save.
            filename (PathLike): The path to the pytorch file to write.
        """
        features = volume.tensor.detach().cpu().contiguous()
        matrix = volume.geometry.tensor.detach().cpu().contiguous()
        torch.save({'v': features, 'm': matrix}, filename)


# enabled volume IO protocol classes
volume_io_protocols = [
    NiftiArrayIO,
    MghArrayIO,
    PytorchVolumeIO,
]



import itertools
import torch
import random
import voxel as vx


def all_orientations() -> list:
    """
    Returns all 48 possible grid orientations in world space.
    """
    data = ['LR', 'PA', 'IS']
    choices = [(item[0], item[1]) for item in data]
    orders = list(itertools.permutations(choices))
    orientations = []
    for order in orders:
        perms = list(itertools.product(*order))
        orientations.extend([''.join(p) for p in perms])
    return orientations


def test_orientation_init() -> None:

    # test construction from a few hardcoded orientation matrices

    assert 'RAS' == vx.Orientation(vx.AffineMatrix(torch.tensor([
        [1, 0, 0],
        [0, 1, 0],
        [0, 0, 1]])))

    assert 'ILA' == vx.Orientation(vx.AffineMatrix(torch.tensor([
        [ 0, -1, 0],
        [ 0,  0, 1],
        [-1,  0, 0]])))

    assert 'PSR' == vx.Orientation(vx.AffineMatrix(torch.tensor([
        [ 0,  0, 1],
        [-1,  0, 0],
        [ 0,  1, 0]])))


def test_orientation_equality() -> None:

    # test that all orientations are equal to themselves
    for name in all_orientations():
        orientation = vx.Orientation(name)
        assert orientation.name == name
        assert orientation == vx.Orientation(name)
        assert orientation == name
        assert name == orientation
    
    # double check that inequality works
    assert vx.Orientation('RAS') != vx.Orientation('LPI')


def test_reorientation() -> None:

    # random volume
    vol = vx.Volume(torch.rand(10, 20, 30), torch.diag(torch.tensor([1, 1.2, 0.8, 1])))

    # cycle through all orientations
    reoriented = vol
    orientations = all_orientations()
    random.shuffle(orientations)
    for name in orientations:

        # reorienting back should give the same original volume
        assert vx.volumes_equal(vol, vol.reorient(name).reorient('RAS'))

        # reorient the volume
        reoriented = reoriented.reorient(name)
        assert reoriented.geometry.orientation == name
    
    # make sure we get back to the original volume
    assert vx.volumes_equal(vol, reoriented.reorient('RAS'))


def test_orientation_slice_direction() -> None:

    shape = (10, 20, 30)
    geometry = vx.AcquisitionGeometry(shape, slice_direction=1)
    vol = vx.Volume(torch.ones(shape), geometry)

    # make sure the explicitly-set slice direction is propagated correctly
    assert vol.reorient('RAS').geometry._explicit_slice_direction == 1
    assert vol.reorient('SRA').geometry._explicit_slice_direction == 2
    assert vol.reorient('PIL').geometry._explicit_slice_direction == 0



"""
Mapping of image voxel coordinates to world space.
"""

from __future__ import annotations

import math
import torch
import voxel as vx


class AcquisitionGeometry(vx.AffineMatrix):
    """
    Geometry representating the linear relationship between volumetric
    image coordinates and a three-dimensional world space.
    """

    def __init__(self,
        baseshape: torch.Size,
        matrix: vx.AffineMatrix | None = None,
        slice_direction: int | None = None,
        device: torch.device | None = None) -> None:
        """
        Args:
            baseshape (Size): The spatial shape (3D dimensions) of the acquisition.
            matrix (AffineMatrix, optional): Affine matrix representing a voxel-to-world
                coordinate transform. If None, it defaults to a shifted identity in
                which the image volume is centered at the world origin.
            slice_direction (int, optional): Voxel-space index representing the dimension
                of slice direction, also known as the through-plane acquisition direction.
                If None, the direction is inferred from the largest voxel spacing.
            device (device, optional): Device to store the matrix on.
        """
        if matrix is None:
            matrix = torch.eye(4, device=device)
            matrix[:3, -1] = -(torch.tensor(baseshape, device=device) - 1) / 2
        super().__init__(matrix, device)
        self._baseshape = torch.Size(baseshape)
        self._explicit_slice_direction = slice_direction

        # the IO 'reference' is used to cache information that is supported by an original image
        # file but not by this class. for example, it is useful for loading and saving file
        # formats that do not explicitly represent the geometry with a voxel-to-world matrix,
        # such as the MGH format. at save-time, if the geometry encoded by this class is similar
        # to that of a cached reference, we instead use the reference header information while
        # saving. this avoids the floating point drift that would otherwise occur during successive
        # reads and writes. this class will propagate any included reference information for
        # functions in which the underlying geometry does not change
        self.reference = {}

    def __setitem__(self, indexing, item):
        """
        Prevents modification of the matrix tensor.
        """
        raise AttributeError('AcquisitionGeometry matrix is read-only')

    @property
    def device(self) -> torch.device:
        """
        Device of the matrix tensor.
        """
        return self.tensor.device

    @property
    def baseshape(self) -> torch.Size:
        """
        Spatial shape (3D dimensions) of the acquisition.
        """
        return self._baseshape

    def _from_new_properties(self,
        baseshape: torch.Size = None,
        matrix: torch.Tensor = None,
        keep_explicit_slice_direction: bool = True,
        keep_reference: bool = False,
        ) -> AcquisitionGeometry:
        """
        A private utility to construct a new geometry with certain propagated information.
        """
        geometry = AcquisitionGeometry(
            baseshape=self.baseshape if baseshape is None else baseshape,
            matrix=self.tensor if matrix is None else matrix,
            slice_direction=self._explicit_slice_direction if keep_explicit_slice_direction else None)

        # shallow copy of the reference cache
        if keep_reference:
            geometry.reference = geometry.reference.copy()

        return geometry

    def _from_tensor_with_new_properties(self, tensor: torch.Tensor) -> AcquisitionGeometry:
        """
        A necessary utility to reimplement when subclassing the AffineMatrix in order to
        support methods such as matrix tensor detaching, device moving, etc.
        """
        return self._from_new_properties(matrix=tensor, keep_reference=True)

    def numel(self) -> int:
        """
        Number of baseshape elements in the acquisition volume.
        """
        return self.baseshape.numel()

    @vx.caching.cached
    def spacing(self) -> torch.Tensor:
        """
        Spacing between voxel centers.
        """
        return torch.linalg.qr(self[:3, :3])[1].diagonal().abs()
 
    @vx.caching.cached
    def slice_direction(self) -> int:
        """
        Index representing the dimension of the slice acquisition
        direction, or through-plane, direction of the acquisition.
        """
        if self._explicit_slice_direction is not None:
            return self._explicit_slice_direction
        else:
            return self.spacing.argmax()

    @property
    def slice_direction_is_explicit(self) -> bool:
        """
        Whether or not the slice direction was explicitly set at
        construction or inferred through heuristics.
        """
        return self._explicit_slice_direction is not None

    @vx.caching.cached
    def in_plane_directions(self) -> list:
        """
        List of two indices representing the dimensions of the in-plane
        acquisition directions, which are orthogonal to the slice direction.
        """
        return [d for d in range(3) if d != self.slice_direction]

    @vx.caching.cached
    def slice_spacing(self) -> torch.Tensor:
        """
        Spacing (separation) between slices.
        """
        return self.spacing[self.slice_direction]

    @vx.caching.cached
    def in_plane_spacing(self) -> torch.Tensor:
        """
        Spacings between voxel centers in the in-plane directions.
        """
        return self.spacing[self.in_plane_directions]

    @vx.caching.cached
    def spacing_ratio(self) -> torch.Tensor:
        """
        Ratio of slice spacing to in-plane spacing.
        """
        return self.slice_spacing / self.in_plane_spacing.mean()

    def is_isotropic(self, rtol: float = 1e-2) -> bool:
        """
        Determine if voxel spacing is isotropic within a relative tolerance.

        Args:
            rtol (float): Relative tolerance of difference across spacings.

        Returns:
            bool: True if isotropic spacing.
        """
        mean = self.in_plane_spacing.mean()
        return self.spacing.allclose(mean, atol=0, rtol=rtol)

    @vx.caching.cached
    def orientation(self) -> Orientation:
        """
        Anatomical orientation of the voxel coordinate system.
        """
        return Orientation(self)

    @vx.caching.cached
    def fov(self) -> torch.Tensor:
        """
        Acquisition field of view in world coordinate space.
        """
        return self.voxel_to_world_units(torch.tensor(self.baseshape, device=self.device))

    def voxel_to_local_coordinates(self, coords: torch.Tensor) -> torch.Tensor:
        """
        Transform voxel coordiates to flipped local grid coordinates in
        the range [-1, 1]. These local coordinates are used for grid sampling in torch.

        Returns:
            torch.Tensor: Transformed local coordinates.
        """
        shape = torch.tensor(self.baseshape).to(coords.device)
        return ((2 * coords + 1) / shape - 1).flip(-1)

    def world_to_voxel_units(self, units: torch.Tensor) -> torch.Tensor:
        """
        Convert world units to voxel units.

        Args:
            units (Tensor): Units of size $(1,)$ or $(3,)$ or $(3, 2)$ in world space.

        Returns:
            Tensor: Units in voxel space.
        """
        units = torch.as_tensor(units, device=self.device).float()
        if units.ndim == 0:
            units = units.repeat(3)
        spacing = self.spacing.unsqueeze(1) if units.ndim == 2 else self.spacing
        return units[self.orientation.dims] / spacing

    def voxel_to_world_units(self, units: torch.Tensor) -> torch.Tensor:
        """
        Convert voxel units to world units.

        Args:
            units (Tensor): Units of size $(1,)$ or $(3,)$ or $(3, 2)$ in voxel space.

        Returns:
            Tensor: Units in world space.
        """
        return (self.spacing * units)[self.orientation.dims.argsort()]

    def conform_units(self,
        units: torch.Tensor,
        source: vx.Space,
        target: vx.Space,
        num: int = None) -> torch.Tensor:
        """
        Conform units to the voxel space. If the input space is 'world', the units are
        converted to world space. Otherwise, the units are just conformed to length 3.

        Args:
            units (Tensor): Units of size $(1,)$ or $(3,)$ or $(3, N)$
            source (Space): Space of the input units.
            target (Space): Space to convert the units to.
            num (int, optional): Number of units $N$ represented in the second dimension
                of the output tensor. If None, the output tensor will have shape $(3,)$.
        
        Returns:
            Tensor: Units of size $(3, N)$ or $(3,)$ in voxel space.
        """
        units = torch.as_tensor(units, device=self.device).float()

        if vx.Space(source) == 'world' and vx.Space(target) == 'voxel':
            units = self.world_to_voxel_units(units)
        elif vx.Space(source) == 'voxel' and vx.Space(target) == 'world':
            units = self.voxel_to_world_units(units)

        return vx.slicing.conform_coordinates(units, num)

    def shift(self, delta: float | torch.Tensor, space: vx.Space) -> AcquisitionGeometry:
        """
        Shift, or translate, the acquisition geometry.

        Args:
            delta (float or Tensor): The shift amount.
            space (Space): The space in which to apply the shift.

        Returns:
            AcquisitionGeometry: The shifted geometry.
        """
        trf = vx.affine.translation_matrix(torch.as_tensor(delta, device=self.device))
        matrix = trf @ self if vx.Space(space) =='world' else self @ trf
        return self._from_new_properties(matrix=matrix)

    def scale(self, factor: float | torch.Tensor, space: vx.Space) -> AcquisitionGeometry:
        """
        Scale the acquisition geometry.

        Args:
            factor (float or torch.Tensor): The scaling factor.
            space (Space): The space in which to apply the scale.

        Returns:
            AcquisitionGeometry: The scaled geometry.
        """
        diag = torch.ones(4, device=self.device)
        diag[:3] = torch.as_tensor(factor, device=self.device)
        trf = torch.diag(diag)
        matrix = trf @ self.tensor if vx.Space(space) =='world' else self.tensor @ trf
        return self._from_new_properties(matrix=matrix)

    def rotate(self,
        rotation: torch.Tensor,
        space: vx.Space,
        corner: bool = False,
        degrees: bool = True) -> AcquisitionGeometry:
        """
        Rotate the acquisition geometry.

        Args:
            rotation (Tensor): Rotation angles (x, y, z). If `degrees` is True, the
                angles are in degrees, otherwise they are in radians.
            space (Space): The space in which to apply the rotation.
            corner (bool, optional): Whether to rotate around the image corner or center. Only
                applicable when the space is 'voxel'. Defaults to True.
            degrees (bool, optional): Whether the angles are defined as degrees or,
                alternatively, as radians.

        Returns:
            AcquisitionGeometry: The rotated geometry.
        """
        rotation = torch.as_tensor(rotation, device=self.device)
        trf = vx.affine.angles_to_rotation_matrix(rotation, degrees=degrees)

        if vx.Space(space) == 'world':
            matrix = trf @ self
        elif not corner:
            center = (torch.tensor(self.baseshape, device=self.device) - 1) / 2
            trf = vx.affine.translation_matrix(center) @ trf @ vx.affine.translation_matrix(-center)
            matrix = self @ trf
        else:
            matrix = self @ trf

        return self._from_new_properties(matrix=matrix)

    def reorient(self, target: vx.Orientation) -> AcquisitionGeometry:
        """
        Reorient the acquisition geometry to a new anatomical voxel orientation.

        Args:
            target (Orientation): Target orientation to reorient to.
        
        Returns:
            AcquisitionGeometry: Reoriented geometry.
        """
        source = self.orientation
        target = cast_orientation(target)

        perm = source.dims.argsort()[target.dims]
        flip = source.flip * target.flip[perm.argsort()]

        baseshape = torch.tensor(self.baseshape)

        trf = vx.AffineMatrix(torch.diag(flip)[:, perm])
        trf[:3, -1] = (baseshape - 1) * (flip < 0)

        slice_direction = None
        if self.slice_direction_is_explicit:
            slice_direction = perm.argsort()[self.slice_direction]

        return AcquisitionGeometry(baseshape[perm], self @ trf.to(self.device), slice_direction=slice_direction)

    def resample(self,
        spacing: float | torch.Tensor = None,
        in_plane_spacing: float | torch.Tensor = None,
        slice_spacing: float | torch.Tensor = None) -> AcquisitionGeometry:
        """
        Resample to a new voxel grid spacing.

        Args:
            spacing (float |Tensor): Target voxel spacing. An isotropic target
                is assumed if a scalar is provided.
            in_plane_spacing (float | Tensor): Target in-plane voxel spacing. Mutually
                exclusive with the `spacing` argument.
            slice_spacing (float | Tensor): Target slice spacing. Mutually exclusive
                except with the `spacing` argument.

        Returns:
            AcquisitionGeometry: Resampled geometry.
        """
        if spacing is None and in_plane_spacing is None and slice_spacing is None:
            raise ValueError('must provide either spacing, in_plane_spacing, or slice_spacing')
        if spacing is not None:
            if in_plane_spacing is not None or slice_spacing is not None:
                raise ValueError('cannot provide spacing with in_plane_spacing or slice_spacing')
        else:
            spacing = self.spacing.clone()
            if in_plane_spacing is not None:
                spacing[self.in_plane_directions] = in_plane_spacing
            if slice_spacing is not None:
                spacing[self.slice_direction] = slice_spacing

        if not torch.is_tensor(spacing):
            spacing = torch.tensor(spacing, dtype=torch.float32)
        if spacing.ndim == 0:
            spacing = spacing.repeat(3)
        if spacing.ndim != 1 or spacing.shape[0] != 3:
            raise ValueError(f'expected 3D spacing, got {spacing.ndim}D')

        # compute new shapes and lengths of the new grid (we'll round up here to avoid losing any signal)
        spacing = spacing.to(self.device)
        curshape = torch.tensor(self.baseshape, dtype=torch.float32, device=spacing.device)
        newshape = (self.spacing * curshape / spacing).ceil().int()

        # determine the new geometry
        scale = spacing / self.spacing
        shift = 0.5 * scale * (1 - newshape / curshape)
        matrix = self.shift(shift, space='voxel').scale(scale, space='voxel')

        # we should only transfer an explicit slice direction if the new spacing was
        # designated directly by a target in-plane and/or slice spacing
        slice_direction = None
        if self.slice_direction_is_explicit and (in_plane_spacing is not None or slice_spacing is not None):
            slice_direction = self.slice_direction

        return vx.AcquisitionGeometry(newshape, matrix, slice_direction=slice_direction)

    def pool(self,
        scale: int = 2,
        spacing_ratio_thresh: float = None) -> AcquisitionGeometry:
        """
        Pool the geometry with a sliding window.

        By default, this will pool over all dimensions, but it can be conditionally
        disabled for the slice dimension based on the ratio of slice vs in-plane spacing, 
        i.e. the value of `geometry.spacing_ratio`. For example, if the slice spacing is
        `spacing_ratio_thresh` times greater than the in-plane spacing, the slice dimension
        will not be pooled. Mind that if the resulting pooled volume has a slice spacing
        less than the in-plane spacing, it will be resampled to an isotropic resolution.

        There is no analogous `unpool` method because there is complexity in determining
        the desired unpooling strategy. To return to the original geometry, instead use
        the `resample_like` method. If no reference geometry is available, just use
        the `reshape` method to upsample.

        Note that this implementation must mirror the pooling operation used by the volume
        class. Any changes to the pooling operation in one class must be reflected in the other.

        Args:
            scale (int, optional): The size of the pooling window. Defaults to 2.
            spacing_ratio_thresh (float, optional): Slice spacing ration that determines
                whether the slice dimension is pooled. This is disabled by default.

        Returns:
            AcquisitionGeometry: Pooled geometry.
        """

        # dowsample factors for each dimension
        factors = [1 if d == 1 else scale for d in self.baseshape]

        # check if we should pool the slice dimension based on the spacing ratio threshold
        slice_dim_pooling = spacing_ratio_thresh is None or self.spacing_ratio < spacing_ratio_thresh

        if not slice_dim_pooling:
            factors[self.geometry.slice_direction] = 1

        pooled_shape = [math.ceil(s / f) for s, f in zip(self.baseshape, factors)]

        # adjust the geometry based on the pooling factors
        shift = [0.5 * (f - 1) for f in factors]
        adjusted = self.shift(shift, space='voxel').scale(factors, space='voxel')
        pooled = vx.AcquisitionGeometry(pooled_shape, adjusted)

        # if the slice dimension was not pooled and the resulting slice spacing
        # is less than the in-plane spacing, we need to resample the slice dimension
        if not slice_dim_pooling and self.spacing_ratio < scale:
            spacing = pooled.geometry.spacing.clone()
            spacing[self.geometry.slice_direction] = spacing[self.geometry.in_plane_directions].mean()
            pooled = pooled.resample(spacing)

        # sanity check
        if pooled.spacing_ratio < 0.99:
            raise ValueError('unexpected spacing ratio after pooling operation')

        return pooled

    def reshape(self, baseshape: torch.Size, from_origin: bool = False) -> AcquisitionGeometry:
        """
        Modify the spatial extent of the volume geometry, cropping or padding around the
        center image to fit a given **baseshape**.

        This method is symmetric in that performing a reverse reshape operation
        will always yeild the original geometry.

        args:
            baseshape (Size): Target spatial (3D) shape.
            from_origin (bool, optional): If True, padding or cropped will be done
                at the ends of the image shape and not centered.
        
        returns:
            AcquisitionGeometry: Reshaped geometry.
        """
        baseshape = torch.Size(baseshape)
        if baseshape == self.baseshape:
            return self

        if from_origin:
            return self._from_new_properties(baseshape=baseshape, keep_reference=True)

        shift = (torch.tensor(self.baseshape) - torch.tensor(baseshape)) // 2
        return self.shift(shift, space='voxel').reshape(baseshape, from_origin=True)

    def pad(self, margin: float | torch.Tensor, space: vx.Space) -> AcquisitionGeometry:
        """
        Pad the spatial extent of the volume geometry by a given margin. Note that
        a negative margin value will result in trimming (cropping).

        args:
            margin (float or Tensor, optional): Delta of specified units to
                pad (or crop) the volume by in each direction. Can be of size
                $(1,)$, $(3,)$, or $(3, 2)$.
            space (Space): The space of the margin, either 'voxel' or 'world'.

        returns:
            AcquisitionGeometry: Reshaped volume geometry.
        """
        margin = self.conform_units(margin, space, 'voxel', 2).cpu().round().int()
        new_shape = torch.tensor(self.baseshape) + margin.sum(-1)
        return self.shift(-margin[:, 0], space='voxel').reshape(new_shape, from_origin=True)

    def trim(self, margin: float | torch.Tensor, space: vx.Space) -> AcquisitionGeometry:
        """
        Trim the spatial extent of the volume geometry by a given margin. This is
        equivalent to padding with negative margin values.

        args:
            margin (float or Tensor, optional): Delta of specified units to
                trim the volume by in each direction. Can be of size $(1,)$,
                $(3,)$, or $(3, 2)$.
            space (Space): The space of the margin, either 'voxel' or 'world'.

        returns:
            AcquisitionGeometry: Reshaped volume geometry.
        """
        return self.pad(-margin, space)

    def bounds(self,
        margin: float | torch.Tensor = None,
        space: vx.Space = 'world') -> vx.Mesh:
        """
        Compute a box mesh enclosing the bounds of the grid.

        Args:
            margin (float or Tensor, optional): Margin to expand the cropping boundary.
                Can be a positive or negative delta.
            space (Space, optional): Space of the margin values, either 'voxel' or 'world'.

        Returns:
            Mesh: Bounding box mesh in world-space coordinates.
        """
        minc = torch.zeros(3, device=self.device)
        maxc = torch.tensor(self.baseshape, device=self.device).float() - 1
        
        # expand (or shrink) margin around border
        if margin is not None:
            margin = self.conform_units(margin, space, 'voxel', 2)
            minc -= margin[:, 0]
            maxc += margin[:, 1]

        # build the world-space bounding box mesh
        mesh = vx.mesh.construct_box_mesh(minc, maxc)
        return mesh.transform(self)

    def fit_to_bounds(self,
        bounds: vx.Mesh,
        margin: float | torch.Tensor = None,
        space: vx.Space = 'world') -> AcquisitionGeometry:
        """
        Fit the geometry to the bounds of a world-space mesh.

        Args:
            bounds (Mesh): Mesh defining the bounds to fit to.
            margin (float or Tensor, optional): Margin to expand the bounding box
                around the bounds. Can be a positive or negative delta.
            space (Space, optional): Space of the margin values, either 'voxel' or 'world'.

        Returns:
            AcquisitionGeometry: Reshaped geometry.
        """
        points = self.inverse().transform(bounds.vertices.detach())
        minc = points.amin(0).floor().int()
        maxc = points.amax(0).ceil().int() + 1

        if margin is not None:
            margin = self.conform_units(margin, space, 'voxel', 2)
            minc -= margin[:, 0]
            maxc += margin[:, 1]

        return self.shift(minc, 'voxel').reshape(maxc - minc, from_origin=True)

    def zeros_like(self,
        channels: int = 1,
        dtype: torch.dtype | None = None) -> vx.Volume:
        """
        Create a volume of zeros.

        Args:
            channels (int, optional): Number of channels for the new volume.
            dtype (torch.dtype, optional): Target data type.

        Returns:
            Volume: A new volume instance filled with zeros.
        """
        shape = (channels, *self.baseshape)
        return vx.Volume(torch.zeros(shape, dtype=dtype, device=self.device), self)

    def ones_like(self,
        channels: int = 1,
        dtype: torch.dtype | None = None) -> vx.Volume:
        """
        Create a volume of ones.

        Args:
            channels (int, optional): Number of channels for the new volume.
            dtype (torch.dtype, optional): Target data type.

        Returns:
            Volume: A new volume instance filled with ones.
        """
        shape = (channels, *self.baseshape)
        return vx.Volume(torch.ones(shape, dtype=dtype, device=self.device), self)

    def full_like(self,
        fill: float,
        channels: int = 1,
        dtype: torch.dtype | None = None) -> vx.Volume:
        """
        Create a volume filled with a specific value.

        Args:
            fill (float): The fill value.
            channels (int, optional): Number of channels for the new volume.
            dtype (torch.dtype, optional): Target data type.

        Returns:
            Volume: A new filled volume instance.
        """
        shape = (channels, *self.baseshape)
        return vx.Volume(torch.full(shape, fill, dtype=dtype, device=self.device), self)

    def rand_like(self,
        channels: int = 1,
        dtype: torch.dtype | None = None) -> vx.Volume:
        """
        Create a volume of random values. Values are sampled from a uniform
        distribution on the interval [0, 1).

        Args:
            channels (int, optional): Number of channels for the new volume.
            dtype (torch.dtype, optional): Target data type.

        Returns:
            Volume: A new random volume instance.
        """
        shape = (channels, *self.baseshape)
        return vx.Volume(torch.rand(shape, dtype=dtype, device=self.device), self)

    def randn_like(self,
        channels: int = 1,
        dtype: torch.dtype | None = None) -> vx.Volume:
        """
        Create a volume of random values. Values are sampled from a normal
        distribution with mean 0 and variance 1.

        Args:
            channels (int, optional): Number of channels for the new volume.
            dtype (torch.dtype, optional): Target data type.

        Returns:
            Volume: A new random volume instance.
        """
        shape = (channels, *self.baseshape)
        return vx.Volume(torch.randn(shape, dtype=dtype, device=self.device), self)


def cast_acquisition_geometry(obj: vx.Volume | AcquisitionGeometry) -> AcquisitionGeometry:
    """
    Cast item to an AcquisitionGeometry

    Args:
        obj (Volume | AcquisitionGeometry): Object to cast.
    
    Returns:
        AcquisitionGeometry
    """
    if isinstance(obj, vx.Volume):
        return obj.geometry
    elif isinstance(obj, vx.AcquisitionGeometry):
        return obj
    else:
        raise ValueError(f'cannot cast {type(obj)} to AcquisitionGeometry')


class Orientation:
    """
    The anatomical orientation of the voxel coordinate system.

    There are 48 possible orientations of voxel data in a 3D grid. These are
    defined by the permutations and flips of the grid axes relative to a global
    coordinate space. By default, we use a 'RAS' anatomical world coordinate system,
    which is characterized by the following axis directions:
    
        - x axis: left (L) to right (R)
        - y axis: posterior (P) to anterior (A)
        - z axis: inferior (I) to superior (S)

    Thus, orientations are defined by a three-character string, which maps a voxel
    coordinate axis to the corresponding anatomical axis direction in 'RAS' space.
    """

    def __init__(self, item: str | vx.AffineMatrix) -> None:
        """
        Args:
            item (str | AffineMatrix): Orientation string or affine matrix.
        """

        # this defines the axes of the anatomical world space. in the future, this
        # could be a global setting configured changed by the user
        self.axes = 'LRPAIS'

        if isinstance(item, str):
            try:
                indices = [self.axes.index(x) for x in item]
            except:
                raise ValueError(f'orientation string must contain one of {self.axes}')
            self.dims = torch.as_tensor([0, 0, 1, 1, 2, 2])[indices]
            self.flip = torch.as_tensor([-1, 1, -1, 1, -1, 1])[indices]

        elif isinstance(item, vx.AffineMatrix):
            tensor = item[:3, :3].detach().cpu()
            # self.dims = tensor.abs().argmax(1)
            # self.flip = tensor[tensor.abs().argmax(0), [0, 1, 2]].sign().int()
            self.dims = tensor.abs().argmax(0)
            self.flip = tensor[self.dims, [0, 1, 2]].sign().int()

        else:
            return ValueError(f'cannot create orientation from {type(item)}')

    @property
    def name(self) -> str:
        """
        Abbreviated orientation name.
        """
        indices = self.dims * 2 + (self.flip > 0)
        return ''.join([self.axes[i] for i in indices])

    def __repr__(self) -> str:
        return f"Orientation('{self.name}')"
    
    def __eq__(self, other: Orientation) -> bool:
        other = cast_orientation(other)
        return torch.equal(self.dims, other.dims) and torch.equal(self.flip, other.flip)

    def dim_map(self, target: Orientation) -> torch.Tensor:
        """
        Map dimensions from this orientation to another.

        Args:
            target (Orientation): Target orientation.

        Returns:
            Tensor: 3D integer mapping of dimensions.
        """
        return self.dims.argsort()[cast_orientation(target).dims]

    def view(self, dim: int) -> str:
        """
        Get the name of the view plane (axial, coronal, sagittal) for a given dimension.

        Args:
            dim (int): Dimension index.
        
        Returns:
            str: View plane name.
        """
        views = ['axial', 'coronal', 'sagittal']
        return views[vx.Orientation('SAR').dim_map(self)[dim]]


def cast_orientation(obj) -> Orientation:
    """
    Cast object to an Orientation instance.

    Args:
        obj (any): Object to cast.
    
    Returns:
        Orientation: Casted orientation.
    """
    if isinstance(obj, Orientation):
        return obj
    return Orientation(obj)



"""
Methods related to file reading and writing.
"""

from __future__ import annotations

import os
import pathlib
import numpy as np
import torch


def check_file_readability(filename: os.PathLike) -> None:
    """
    Raise an exception if a file cannot be read.

    Args:
        filename (PathLike): Path to file.
    """
    if not isinstance(filename, pathlib.Path):
        filename = pathlib.Path(filename)

    if filename.is_dir():
        raise ValueError(f'{filename} is a directory, not a file')

    if not filename.is_file():
        raise FileNotFoundError(f'{filename} is not a file')

    if not os.access(filename, os.R_OK):
        raise PermissionError(f'{filename} is not a readable file')


class IOProtocol:
    """
    Abstract (and private) protocol class to implement filetype-specific reading and writing.

    Subclasses must override the `load` and `save` methods, and set the `name` and `extensions`
    global class members.
    """
    name = ''
    extensions = []

    @classmethod
    def primary_extension(cls) -> str:
        """
        Return the primary (first) file extension of the protocol.

        Returns:
            str: The primary file extension.
        """
        if not cls.extensions:
            return ''
        elif isinstance(cls.extensions, str):
            return cls.extensions
        return cls.extensions[0]

    @classmethod
    def enforce_extension(cls, filename: os.PathLike) -> pathlib.Path:
        """
        Enforce a valid protocol extension on a filename. Returns the corrected filename.

        Args:
            filename (PathLike): The filename to enforce the extension on.

        Returns:
            Path: The filename with the enforced extension.
        """
        if str(filename).lower().endswith(cls.extensions):
            return filename
        return pathlib.Path(filename).with_suffix(cls.primary_extension())

    def load(self, filename: os.PathLike) -> object:
        """
        File load function to be implemented for each subclass.

        Args:
            filename (PathLike): The filename to load.

        Raises:
            NotImplementedError: If the subclass does not implement this method.
        """
        raise NotImplemented(f'reading file format of {os.path.basename(filename)} is not implemented yet')

    def save(self, obj: object, filename: os.PathLike) -> None:
        """
        File save function to be implemented for each subclass.

        Args:
            obj (object): The object to save.
            filename (PathLike): The filename to save the object to.

        Raises:
            NotImplementedError: If the subclass does not implement this method.
        """
        raise NotImplemented(f'writing file format of {os.path.basename(filename)} is not implemented yet')


def find_protocol_by_name(protocols: list, fmt: str) -> IOProtocol | None:
    """
    Find IO protocol by format name.

    Args:
        protocols (list of IOProtocol): List of IOProtocol classes to search.
        fmt (str): File format name.

    Returns:
        IOProtocol: Matched IO protocol class or None if not found.
    """
    fmt = fmt.lower()
    return next((p for p in protocols if fmt == p.name), None)


def find_protocol_by_extension(protocols: list, filename: os.PathLike) -> IOProtocol | None:
    """
    Find IO protocol by extension type.

    Args:
        protocols (list of IOProtocol): List of IOProtocol classes to search.
        filename (PathLike): Filename to grab extension of.

    Returns:
        IOProtocol: Matched IO protocol class or None if not found.
    """
    lowercase = str(filename).lower()
    return next((p for p in protocols if lowercase.endswith(p.extensions)), None)


def get_all_extensions(protocols: list) -> list:
    """
    Returns all extensions in a list of protocols.

    Args:
        protocols (list of str): List of IOProtocol classes to search.

    Returns:
        list of str: List of extensions.
    """
    extensions = []
    for protocol in protocols:
        if isinstance(protocol.extensions, str):
            extensions.append(protocol.extensions)
        else:
            extensions.extend(protocol.extensions)
    return extensions


def numpy_to_tensor(x, dtype: torch.dtype = None, copy: bool = False) -> torch.Tensor:
    """
    Safely convert a numpy array to a tensor.

    Args:
        x: Numpy ndarray or array-like to convert
        dtype (dtype, optional): Optional torch dtype
        copy (bool, optional): Force a copy even if sharing is possible

    Returns:
        Tensor: Converted tensor
    """
    if isinstance(x, torch.Tensor):
        return x

    if not isinstance(x, np.ndarray):
        x = np.asarray(x)

    # only numeric/boolean dtypes are meaningful for tensors
    if x.dtype.kind not in "biufc?":  # bool/int/uint/float/complex/bool
        raise TypeError(f"unsupported dtype kind '{x.dtype.kind}' for tensor conversion")

    # normalize endianness: make sure dtype is native (or non-endian, e.g. '|u1')
    # np.dtype.byteorder: '=' native, '<' little, '>' big, '|' not applicable
    bo = x.dtype.byteorder
    if bo not in ('=', '|'):
        # if platform is little-endian and x is big-endian (or vice versa), swap
        if (bo == '>' and np.little_endian) or (bo == '<' and not np.little_endian):
            # byteswap returns a view by default - newbyteorder('=') sets native
            x = x.byteswap().view(x.dtype.newbyteorder('='))

    # if requested, or if array is read-only, make a copy so torch can safely own/write
    if copy or (hasattr(x, "flags") and not x.flags.writeable):
        x = x.copy()

    # use as_tensor to share memory when possible but copy when required
    return torch.as_tensor(x, dtype=dtype)
